from unittest import TestCase

from musicxml.xmlelement.xmlelement import *


class TestChoice(TestCase):
    def test_xml_element_with_choice(self):
        xml_encoding = XMLEncoding()
        xml_encoding.add_child(XMLEncoder('bla'))
        xml_encoding.add_child(XMLEncoder('blue'))

        assert len(xml_encoding.get_children()) == 2



import copy
from contextlib import redirect_stdout
from pathlib import Path
from string import Template
import xml.etree.ElementTree as ET

from musicxml.generate_classes.utils import get_complex_type_all_base_classes, get_all_et_elements, musicxml_xsd_et_root
from musicxml.xsd.xsdtree import XSDTree, XSD_TREE_DICT
from musicxml.xsd.xsdsimpletype import *

sources_path = Path(__file__).parent / 'musicxml_4_0.xsd'
default_path = Path(__file__).parent / 'defaults' / 'xsdcomplextype.py'
target_path = Path(__file__).parent.parent / 'xsd' / 'xsdcomplextype.py'

template_string = """
class $class_name($base_classes):
    \"\"\"$doc\"\"\"
    
    _SIMPLE_CONTENT = $simple_content
    _XSD_TREE = XSD_TREE_DICT['complexType']['$name']
"""

xsd_complex_types = ['XSDComplexType', 'XSDComplexTypeScorePartwise', 'XSDComplexTypePart', 'XSDComplexTypeMeasure',
                     'XSDComplexTypeDirective', 'XSDComplexTypeNote']


def complex_type_class_as_string(complex_type):
    def get_doc():
        output = xsd_tree.get_doc()
        if simple_content:
            simple_doc = eval(simple_content).get_xsd_tree().get_doc()
            if simple_doc and simple_doc != "":
                if output and output != "":
                    output += '\n'
                    output += '\n'
                output += '``simpleContent``: '
                output += simple_doc
        if not output:
            output = ""
        return output

    name = complex_type[0]
    xsd_tree = complex_type[1]
    class_name = xsd_tree.xsd_element_class_name
    xsd_complex_types.append(class_name)
    base_class_names = []
    simple_content = None
    for cls_name in get_complex_type_all_base_classes(xsd_tree):
        if cls_name.startswith('XSDSimpleType'):
            if simple_content is not None:
                raise NotImplementedError('More than one Simple Type as base class.')
            simple_content = cls_name
        else:
            base_class_names.append(cls_name)

    t = Template(template_string).substitute(class_name=class_name, base_classes=', '.join(base_class_names), simple_content=simple_content,
                                             doc=get_doc(), name=name)
    return t


# all_complex_type_et_elements = [ct for ct in get_all_et_elements(sources_path, 'complexType') if ct.attrib['name'] != 'note']

# all_complex_type_names = [ct.attrib.get('name') for ct in musicxml_xsd_et_root.findall(
#     f".//{{*}}complexType") if ct.attrib.get('name')]

# all_complex_type_names.remove('note')

all_complex_type_et_elements = XSD_TREE_DICT['complexType'].copy()
all_complex_type_et_elements.pop('note')

with open(target_path, 'w+') as f:
    with open(default_path, 'r') as default:
        with redirect_stdout(f):
            print(default.read())
    with redirect_stdout(f):
        for complex_type in all_complex_type_et_elements.items():
            print(complex_type_class_as_string(complex_type))
        print(f'__all__={xsd_complex_types}')



import cProfile
import datetime
from contextlib import redirect_stdout
from pathlib import Path

from musicxml.parser.parser import parse_musicxml


def p():
    start_reading = datetime.datetime.now()
    print("start reading score")
    score = parse_musicxml(Path(__file__).parent / 'parser_test.xml')
    start_writing = datetime.datetime.now()
    print(f"start writing score :{start_writing - start_reading}")
    score.write(Path(__file__).parent / 'parser_test_create.xml')
    end = datetime.datetime.now()
    print(f"end writing score :{end - start_reading}")


# with open(Path(__file__).with_suffix('.txt'), '+w') as f:
#     with redirect_stdout(f):
#         cProfile.run('p()', sort="tottime")



import re
import xml.etree.ElementTree as ET
from typing import Any, Optional

from musicxml.util.core import get_cleaned_token
from musicxml.xsd.xsdtree import XSDTree, XSDTreeElement, XSD_TREE_DICT


class XSDSimpleType(XSDTreeElement):
    """
    Parent Class for all SimpleType classes
    """
    _TYPES: list[type] = []
    _UNION: list[Any] = []
    _FORCED_PERMITTED: list[str] = []
    _PERMITTED: list[str] = []
    _PATTERN: Optional[str] = None

    def __init__(self, value: Any, parent=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent
        if not self._PERMITTED:
            self._populate_permitted()
        if not self._FORCED_PERMITTED:
            self._populate_forced_permitted()
        if self._UNION:
            self._TYPES = []
            for t_ in self._UNION:
                self._TYPES.extend(t_._TYPES)
        self._populate_pattern()
        self._value = None
        self.value = value

    def _check_value(self, v):
        if self._UNION:
            errors = []
            for t_ in self._UNION:
                try:
                    t_(v)
                    return
                except TypeError:
                    pass
                except ValueError as err:
                    errors.append(err.args[0])
            raise ValueError(self._get_error_class(), errors)

        elif v in self._FORCED_PERMITTED:
            return
        if self._PERMITTED:
            if v not in self._PERMITTED:
                raise ValueError(f"{self._get_error_class()}.value '{v}' must be in {self._PERMITTED}")
        elif self._PATTERN:
            restriction = self.get_xsd_tree().get_restriction()
            if restriction:
                if restriction.get_attributes()['base'] == 'xs:date':
                    XSDSimpleTypeDate(v)
                elif restriction.get_attributes()['base'] == 'xs:token':
                    v = XSDSimpleTypeToken(v).value
                elif restriction.get_attributes()['base'] == 'xs:smufl-glyph-name':
                    XSDSimpleTypeSmuflGlyphName(v)
            if re.compile(self._PATTERN).fullmatch(v) is None:
                raise ValueError(
                    f"{self._get_error_class()}.value '{v}' must match the following pattern: {self._PATTERN}")
        else:
            restriction = self.get_xsd_tree().get_restriction()
            if restriction:
                restriction_children = restriction.get_children()
                for child in restriction_children:
                    if child.tag == 'minLength' and len(v) < int(child.get_attributes()['value']):
                        raise ValueError(
                            f"{self._get_error_class()}.value '{v}' must have a length >= 1")
                    if child.tag == 'minExclusive' and v <= int(child.get_attributes()['value']):
                        raise ValueError(
                            f"{self._get_error_class()}.value '{v}' must be greater than"
                            f" '{child.get_attributes()['value']}'")
                    if child.tag == 'minInclusive' and v < int(child.get_attributes()['value']):
                        raise ValueError(
                            f"{self._get_error_class()}.value '{v}' must be greater than or equal to"
                            f" '{child.get_attributes()['value']}'")
                    if child.tag == 'maxInclusive' and v > int(child.get_attributes()['value']):
                        raise ValueError(
                            f"{self._get_error_class()}.value '{v}' must be less than or equal to"
                            f" '{child.get_attributes()['value']}'")

    def _check_value_type(self, value):
        if value in self._FORCED_PERMITTED:
            return
        if isinstance(self._TYPES, str):
            raise TypeError
        if self._TYPES == str or not hasattr(self._TYPES, '__iter__'):
            raise TypeError

        if True in [isinstance(value, type_) for type_ in self._TYPES]:
            pass
        else:
            message = f"{self._get_error_class()}'s value '{value}' can only be of types {[type_.__name__ for type_ in self._TYPES]} not {type(value).__name__}."
            if self._PERMITTED:
                message += f" {self._get_error_class()}.value must in {self._PERMITTED}"
            if self._FORCED_PERMITTED:
                message += f" {self._get_error_class()}.value can also be {self._FORCED_PERMITTED}"
            raise TypeError(message)

    def _get_error_class(self):
        if self.parent:
            return self.parent.__class__.__name__
        else:
            return self.__class__.__name__

    def _populate_permitted(self):
        self._PERMITTED = self.get_xsd_tree().get_permitted()

    def _populate_forced_permitted(self):
        union = self.get_xsd_tree().get_union()
        if union and union.get_children() and union.get_children()[0].tag == 'simpleType':
            intern_simple_type = union.get_children()[0]
            enumerations = [child for child in intern_simple_type.get_restriction().get_children() if child.tag
                            == 'enumeration']
            self._FORCED_PERMITTED = [enumeration.get_attributes()['value'] for enumeration in enumerations]

    def _populate_pattern(self):
        pattern = self.get_xsd_tree().get_pattern(self.__class__.__mro__[1].get_xsd_tree())
        if pattern:
            self._PATTERN = pattern

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._check_value_type(v)
        if v not in self._FORCED_PERMITTED:
            self._check_value(v)
        self._value = v

    def __repr__(self):
        return str(self.value)


class XSDSimpleTypeInteger(XSDSimpleType):
    _TYPES = [int]
    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="integer" id="integer">
            <xs:restriction base="xs:decimal">
                <xs:fractionDigits value="0" fixed="true"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, v):
        self._check_value_type(v)
        super(XSDSimpleTypeInteger, type(self)).value.fset(self, v)


class XSDSimpleTypeNonNegativeInteger(XSDSimpleTypeInteger):
    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="nonNegativeInteger" id="nonNegativeInteger">
            <xs:restriction base="xs:integer">
                <xs:minInclusive value="0"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, v):
        super(XSDSimpleTypeNonNegativeInteger, type(self)).value.fset(self, v)
        if v < 0:
            raise ValueError(f'value {v} must be non negative.')


class XSDSimpleTypePositiveInteger(XSDSimpleTypeInteger):
    _TYPES = [int]
    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="positiveInteger" id="positiveInteger">
            <xs:restriction base="xs:nonNegativeInteger">
                <xs:minInclusive value="1"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, v):
        super(XSDSimpleTypePositiveInteger, type(self)).value.fset(self, v)
        try:
            if v <= 0:
                raise ValueError(f'value {v} must be greater than 0.')
        except TypeError:
            # Important because of XSDSimpleTypePositiveIntegerOrEmpty
            pass


class XSDSimpleTypeDecimal(XSDSimpleType):
    _TYPES = [float, int]
    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="decimal" id="decimal">
            <xs:restriction base="xs:anySimpleType">
                <xs:whiteSpace value="collapse" fixed="true"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, v):
        self._check_value_type(v)
        super(XSDSimpleTypeDecimal, type(self)).value.fset(self, v)


class XSDSimpleTypeString(XSDSimpleType):
    _TYPES = [str]
    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="string" id="string">
            <xs:restriction base="xs:anySimpleType">
                <xs:whiteSpace value="preserve"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, v):
        self._check_value_type(v)
        super(XSDSimpleTypeString, type(self)).value.fset(self, v)


class XSDSimpleTypeToken(XSDSimpleTypeString):
    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="token" id="token">
            <xs:restriction base="xs:normalizedString">
                <xs:whiteSpace value="collapse"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, v):
        super(XSDSimpleTypeToken, type(self)).value.fset(self, v)
        v = get_cleaned_token(v)
        self._value = v


class XSDSimpleTypeDate(XSDSimpleTypeString):
    # [-]CCYY-MM-DD[Z|(+|-)hh:mm]
    # https://www.oreilly.com/library/view/regular-expressions-cookbook/9781449327453/ch04s07.html

    _XSD_TREE = XSDTree(ET.fromstring(
        """
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="date" id="date">
            <xs:restriction base="xs:anySimpleType">
                <xs:whiteSpace value="collapse" fixed="true"/>
            </xs:restriction>
        </xs:simpleType>
        """
    ))
    _PATTERN = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])(Z|[+-](?:2[0-3]|[01][0-9]):[' \
               r'0-5][0-9])?$'

# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_simple_types.py
# -----------------------------------------------------



# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_xml_elements.py
# -----------------------------------------------------
import copy
import xml.etree.ElementTree as ET
from typing import Optional, List, Callable, Union

from musicxml.exceptions import XSDWrongAttribute, XSDAttributeRequiredException, XMLElementChildrenRequired
from musicxml.generate_classes.utils import musicxml_xsd_et_root, ns
from verysimpletree.tree import Tree
from musicxml.util.core import cap_first, replace_key_underline_with_hyphen
from musicxml.xmlelement.containers import containers
from musicxml.xmlelement.exceptions import XMLElementCannotHaveChildrenError
from musicxml.xmlelement.xmlchildcontainer import DuplicationXSDSequence
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdsimpletype import *
from musicxml.xsd.xsdtree import XSDTree, XSD_TREE_DICT


class XMLElement(Tree):
    """
    Parent class of all xml elements.
    """
    _PROPERTIES = {'compact_repr', 'is_leaf', 'attributes', 'child_container_tree', 'possible_children_names',
                   'et_xml_element', 'name', 'type_', 'value_', 'parent_xsd_element', 'xsd_check'}
    TYPE = None
    _SEARCH_FOR_ELEMENT = ''
    XSD_TREE = None

    def __init__(self, value_='', xsd_check=True, **kwargs):
        self._fill_xsd_tree()
        self._type = None
        super().__init__()
        self._xsd_check = None
        self._value_ = None
        self._attributes = {}
        self._et_xml_element = None
        self._child_container_tree = None
        self._unordered_children = []
        self.value_ = value_
        self.xsd_check = xsd_check
        self._set_attributes(kwargs)

        self._create_child_container_tree()

    @classmethod
    def _fill_xsd_tree(cls):
        if cls.XSD_TREE is None:
            cls.XSD_TREE = XSDTree(musicxml_xsd_et_root.find(cls._SEARCH_FOR_ELEMENT))

    def _check_attribute(self, name, value):
        attributes = self.TYPE.get_xsd_attributes()
        allowed_attributes = [attribute.name for attribute in attributes]
        if name not in [attribute.name for attribute in self.TYPE.get_xsd_attributes()]:
            raise XSDWrongAttribute(f"{self.__class__.__name__} has no attribute {name}. Allowed attributes are: {allowed_attributes}")
        for attribute in attributes:
            if attribute.name == name:
                return attribute(value)

    def _check_child_to_be_added(self, child):
        if not isinstance(child, XMLElement):
            raise TypeError

    def _check_required_attributes(self):
        if self.TYPE.get_xsd_tree().is_complex_type:
            required_attributes = [attribute for attribute in self.TYPE.get_xsd_attributes() if attribute.is_required]
            for required_attribute in required_attributes:
                if required_attribute.name not in self.attributes:
                    raise XSDAttributeRequiredException(f"{self.__class__.__name__} requires attribute: {required_attribute.name}")

    def _check_required_value(self):
        if self.TYPE.get_xsd_tree().is_simple_type and self.value_ is None:
            raise ValueError(f"{self.__class__.__name__} needs a value.")

    def _convert_attribute_to_child(self, name, value):
        if not name.startswith('xml_'):
            raise NameError
        child_name = name.replace('xml_', '')

        if '-'.join(child_name.split('_')) not in self.possible_children_names:
            raise NameError

        child_class_name = 'XML' + ''.join([cap_first(partial) for partial in child_name.split('_')])
        child_class = eval(child_class_name)

        found_child = self.find_child(child_class_name)
        if isinstance(value, child_class):
            if found_child:
                self.replace_child(found_child, value)
            else:
                self.add_child(value)
        elif value is None:
            if found_child:
                self.remove(found_child)
        else:
            if found_child:
                found_child.value_ = value
            else:
                self.add_child(child_class(value))

    def _create_child_container_tree(self):
        try:
            if self.TYPE.get_xsd_tree().is_complex_type:
                self._child_container_tree = copy.copy(containers[self.TYPE.__name__])
                self._child_container_tree._parent_xml_element = self
        except KeyError:
            pass

    def _create_et_xml_element(self):
        self._et_xml_element = ET.Element(self.name, {k: str(v) for k, v in self.attributes.items()})
        if self.value_ is not None:
            self._et_xml_element.text = str(self.value_)
        for child in self.get_children():
            self._et_xml_element.append(child.et_xml_element)
        ET.indent(self._et_xml_element, space="  ", level=self.get_level())

    def _final_checks(self, intelligent_choice=False):
        self._check_required_value()
        if self._child_container_tree:
            required_children = self._child_container_tree.get_required_element_names(intelligent_choice=intelligent_choice)
            if required_children:
                raise XMLElementChildrenRequired(f"{self.__class__.__name__} requires at least following children: {required_children}")

        self._check_required_attributes()

        for child in self.get_children():
            child._final_checks(intelligent_choice=intelligent_choice)

    def _get_attributes_error_message(self, wrong_name):
        attributes = self.TYPE.get_xsd_attributes()
        allowed_attributes = [attribute.name for attribute in attributes]
        return f"{self.__class__.__name__} has no attribute {wrong_name}. Allowed attributes are: " \
               f"{sorted(allowed_attributes)} or possible " \
               f"children as attributes: {sorted(['xml_' + '_'.join(ch.split('-')) for ch in self.possible_children_names])}"

    def _set_attributes(self, val):
        if val is None:
            return

        if self.TYPE.get_xsd_tree().is_simple_type:
            if val:
                raise XSDWrongAttribute(f'{self.__class__.__name__} has no attributes.')

        elif not isinstance(val, dict):
            raise TypeError

        new_attributes = replace_key_underline_with_hyphen(dict_=val)
        none_values_dict = {k: v for k, v in new_attributes.items() if v is None}
        for key in none_values_dict:
            new_attributes.pop(key)
            try:
                self.attributes.pop(key)
            except KeyError:
                pass
        for key in new_attributes:
            self._check_attribute(key, new_attributes[key])
        self._attributes = {**self._attributes, **new_attributes}

    @property
    def attributes(self):
        """
        :return: a dictionary of attributes like {'font-family': 'Arial'}

        >>> t = XMLText(value_='hello', font_family = 'Arial')
        >>> t.attributes
        {'font-family': 'Arial'}
        >>> t.to_string()
        <text font-family="Arial">hello</text>
        """

        return self._attributes

    @property
    def child_container_tree(self):
        """
        :return: A :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer` object which is used to manage and control XMLElements children. The nodes of a :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer`
                 have a core content property of types :obj:`~musicxml.xsd.xsdindicator.XSDSequence`, :obj:`~musicxml.xsd.xsdindicator.XSDChoice`, :obj:`~musicxml.xsd.xsdindicator.XSDGroup` or :obj:`~musicxml.xsd.xsdelement.XSDElement`. :obj:`~musicxml.xsd.xsdelement.XSDElement` are the content type of
                 :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer` leaves where one or more XMLElements of a single type (depending on maxOccur attribute of element)
                 can be added to its xml_elements list. An interaction of xsd indicators (sequence, choice and group) with xsd elements
                 makes it possible to add :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement``'s Children in the right order and control all xsd rules which apply to musicxml. A
                 variety of exceptions help user to control the xml structure of the exported file which they are intending to use as a
                 MusicXML formatfile.
        """
        return self._child_container_tree

    @property
    def et_xml_element(self):
        """
        :return:  A xml.etree.ElementTree.Element which is used to write the MusicXML file.
        """
        self._create_et_xml_element()
        return self._et_xml_element

    @property
    def name(self):
        return self.XSD_TREE.name

    @property
    def possible_children_names(self):
        if not self.child_container_tree:
            return {}
        else:
            return {leaf.content.name for leaf in self.child_container_tree.iterate_leaves()}

    @property
    def value_(self):
        """
        :return: A validated value of :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement`` which will be translated to its text in xml format.
        """
        return self._value

    @value_.setter
    def value_(self, val):
        """
        :param val: Value to be validated and added to :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement``. This value will be translated to xml element's text in xml format.
        """
        self.TYPE(val, parent=self)
        self._value = val

    @classmethod
    def get_xsd(cls):
        """
        :return: Snippet of MusicXML xsd file which is relevant for this :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement``.
        """
        return cls.XSD_TREE.get_xsd()

    @property
    def xsd_check(self) -> bool:
        """
        Set and get xsd_check attribute. Default is ``True``. If set to false methods add_child() and to_string() run no xsd checking.
        :return: bool
        """
        return self._xsd_check

    @xsd_check.setter
    def xsd_check(self, val):
        self._xsd_check = val

    def add_child(self, child: 'XMLElement', forward: Optional[int] = None) -> 'XMLElement':
        """
        :param child: chlild of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement` to be added to :obj:`~musicxml.xmlelement.xmlelement.XMLElement`'s :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer` and _unordered_children.
        :param forward: If there are more than one :obj:`~musicxml.xsd.xsdelement.XSDElement` leaves in :obj:`~child_container_tree`, ``forward`` can be used to determine manually which of these equivocal ``xsd elements`` is going to be used to attach the child.
        :return: Added child.
        """
        if self.xsd_check:
            if not self._child_container_tree:
                raise XMLElementCannotHaveChildrenError()
            self._child_container_tree.add_element(child, forward)
        self._unordered_children.append(child)
        child._parent = self
        return child

    def get_children(self, ordered: bool = True) -> List['XMLElement']:
        """
        :param bool ordered: True or False.
        :return: :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement`` added children. If ordered is False the _unordered_children is returned as a more light weighted way of
                 getting children instead of using the leaves of :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer`.
        """
        if ordered is False or self.xsd_check is False:
            return self._unordered_children
        if self._child_container_tree:
            return [xml_element for leaf in self._child_container_tree.iterate_leaves() for xml_element in leaf.content.xml_elements if
                    leaf.content.xml_elements]
        else:
            return []

    def find_child(self, name: Union['XMLElement', str], ordered: bool = False) -> 'XMLElement':
        """
        :param name: :obj:`~musicxml.xmlelement.xmlelement.XMLElement` child or it's name as string.
        :param ordered: :obj:`~get_children` ordered mode to be used to find first appearance of child.
        :return: found child of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement`.
        """
        if isinstance(name, type):
            name = name.__name__
        for ch in self.get_children(ordered=ordered):
            if ch.__class__.__name__ == name:
                return ch

    def find_children(self, name: Union['XMLElement', str], ordered: bool = False) -> List['XMLElement']:
        """
        :param name: A child of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement` or it's name as string.
        :param ordered: :obj:`~get_children` ordered mode to be used to find children.
        :return: found children of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement`.
        """
        if isinstance(name, type):
            name = name.__name__
        return [ch for ch in self.get_children(ordered=ordered) if ch.__class__.__name__ == name]

    def remove(self, child: 'XMLElement') -> None:
        """
        :param child: child of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement` to be removed. This method must be used to remove a child properly from :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer` and reset its behaviour.
        :return: None
        """

        def remove_duplictation():
            for node in parent_container.reversed_path_to_root():
                if node.up:
                    if isinstance(node.up.content, DuplicationXSDSequence) and len(node.up.get_children()) > 1:
                        remove_duplicate = False
                        for leaf in node.iterate_leaves():
                            if leaf != parent_container and leaf.content.xml_elements:
                                break
                            remove_duplicate = True
                        if remove_duplicate:
                            node.up.remove(node)

        self._unordered_children.remove(child)

        if self.xsd_check:
            parent_container = child.parent_xsd_element.parent_container.get_parent()
            if parent_container.chosen_child == child.parent_xsd_element.parent_container:
                parent_container.chosen_child = None
                parent_container.requirements_fulfilled = False
            child.parent_xsd_element.xml_elements.remove(child)
            child.parent_xsd_element = None
            remove_duplictation()

        child._parent = None
        del child

    def replace_child(self, old: Union['XMLElement', Callable], new: 'XMLElement', index: int = 0) -> 'XMLElement':
        """
        :param old: A child of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement` or a function which is used to find a child to be replaced.
        :param new: Child of type :obj:`~musicxml.xmlelement.xmlelement.XMLElement` to be replaced with.
        :param int index: index of old in list of old appearances
        :return: new xml element
        :rtype: :obj:`~musicxml.xmlelement.xmlelement.XMLElement`
        """
        if hasattr(old, '__call__'):
            list_of_olds = [ch for ch in self.get_children(ordered=True) if old(ch)]
        else:
            list_of_olds = [ch for ch in self.get_children(ordered=True) if ch == old]

        if not list_of_olds:
            raise ValueError(f"{old} not in list.")
        self._check_child_to_be_added(new)
        old_index = self._unordered_children.index(list_of_olds[index])
        old_child = self._unordered_children[old_index]
        self._unordered_children.remove(old_child)
        self._unordered_children.insert(old_index, new)

        if self.xsd_check:
            parent_xsd_element = old_child.parent_xsd_element
            new.parent_xsd_element = parent_xsd_element
            parent_xsd_element._xml_elements = [new if el == old_child else el for el in parent_xsd_element.xml_elements]
        new._parent = self
        old._parent = None
        return new

    def to_string(self, intelligent_choice: bool = False) -> str:
        """
        :param bool intelligent_choice: Set to ``True`` if you wish to use intelligent choice in final checks to be able to change the attachment order of :obj:`~musicxml.xmlelement.xmlelement.XMLElement` children in :obj:`~child_container_tree` if an ``Exception`` was raised and other choices can still be checked. (NO GUARANTEE!)
        :return: String in xml format.
        """
        if self.xsd_check:
            self._final_checks(intelligent_choice=intelligent_choice)
        self._create_et_xml_element()

        return ET.tostring(self.et_xml_element, encoding='unicode') + '\n'

    def __setattr__(self, key, value):
        if key[0] == '_' or key in self._PROPERTIES:
            super().__setattr__(key, value)
        elif key.startswith('xml_'):
            try:
                self._convert_attribute_to_child(name=key, value=value)
            except NameError:
                raise AttributeError(self._get_attributes_error_message(key))
        else:
            try:
                self._set_attributes({key: value})
            except XSDWrongAttribute:
                raise AttributeError(self._get_attributes_error_message(key))

    def __getattr__(self, item):
        try:
            return self.attributes['-'.join(item.split('_'))]
        except KeyError:
            attributes = self.TYPE.get_xsd_attributes()
            allowed_attributes = ['_'.join(attribute.name.split('-')) for attribute in attributes]
            if item in allowed_attributes:
                return None
            else:
                if item.startswith('xml'):
                    child_name = item.replace('xml_', '')
                    for child in self.get_children(ordered=False):
                        if child.name == '-'.join(child_name.split('_')):
                            return child
                    if '-'.join(child_name.split('_')) in self.possible_children_names:
                        return None
                raise AttributeError(self._get_attributes_error_message(item))

# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_xml_elements.py
# -----------------------------------------------------



import copy
from unittest import TestCase

from musicxml.util.core import convert_to_xml_class_name, show_force_valid
from musicxml.xmlelement.exceptions import XMLChildContainerWrongElementError, XMLChildContainerMaxOccursError, \
    XMLChildContainerChoiceHasAnotherChosenChild, XMLChildContainerFactoryError
from musicxml.xmlelement.xmlchildcontainer import XMLChildContainer, DuplicationXSDSequence, XMLChildContainerFactory
from musicxml.xmlelement.xmlelement import *
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdelement import XSDElement
from musicxml.xsd.xsdindicator import *
from musicxml.xsd.xsdtree import XSDTree
import xml.etree.ElementTree as ET


class TestChildContainer(TestCase):
    def setUp(self) -> None:
        element_xsd = '<xs:element xmlns:xs="http://www.w3.org/2001/XMLSchema" name="alter" type="semitones" minOccurs="0"/>'
        sequence_xsd = """
                <xs:sequence xmlns:xs="http://www.w3.org/2001/XMLSchema">
                    <xs:element name="step" type="step"/>
                    <xs:element name="alter" type="semitones" minOccurs="0"/>
                    <xs:element name="octave" type="octave"/>
                </xs:sequence>
        """
        choice_xsd = """<xs:choice xmlns:xs="http://www.w3.org/2001/XMLSchema">
                <xs:element name="pitch" type="pitch"/>
                <xs:element name="unpitched" type="unpitched"/>
                <xs:element name="rest" type="rest"/>
            </xs:choice>
        """
        self.element = XSDElement(XSDTree(ET.fromstring(element_xsd)))
        self.sequence = XSDSequence(XSDTree(ET.fromstring(sequence_xsd)))
        self.choice = XSDChoice(XSDTree(ET.fromstring(choice_xsd)))

    def test_children_container_node_type(self):
        """
        Test that an XMLChildContainer must initiate with an XSDElement, XSDGroup, XSDSequence or
        XSDChoice instances.
        """

        XMLChildContainer(self.element)
        XMLChildContainer(self.sequence)
        XMLChildContainer(self.choice)
        XMLChildContainer(XSDGroupEditorial())
        with self.assertRaises(TypeError):
            XMLChildContainer()
        with self.assertRaises(TypeError):
            XMLChildContainer(3)

    def test_children_container_reset_frozen(self):
        container = XMLChildContainer(self.element)
        container._reset_iterators()

    def test_min_max_occurrences(self):
        container = XMLChildContainer(self.element)
        assert container.min_occurrences == 1
        assert container.max_occurrences == 1
        container = XMLChildContainer(self.element, min_occurrences='0')
        assert container.min_occurrences == 0
        container = XMLChildContainer(self.element, max_occurrences='2')
        assert container.max_occurrences == 2
        container = XMLChildContainer(self.element, max_occurrences='unbounded')
        assert container.max_occurrences == 'unbounded'

    def test_that_all_complex_types_can_create_child_container(self):
        from musicxml.xsd.xsdcomplextype import __all__
        complex_types_with_child_container = []
        complex_types_without_child_container = []
        for complex_type_name in __all__[1:]:
            complex_type = eval(complex_type_name)
            try:
                XMLChildContainerFactory(complex_type).get_child_container()
                complex_types_with_child_container.append(complex_type_name)
            except XMLChildContainerFactoryError:
                complex_types_without_child_container.append(complex_type_name)
                # print(complex_type.get_xsd())

        assert set(complex_types_with_child_container) == {'XSDComplexTypeDynamics', 'XSDComplexTypeMidiInstrument',
                                                           'XSDComplexTypeNameDisplay', 'XSDComplexTypePlay', 'XSDComplexTypeAttributes',
                                                           'XSDComplexTypeBeatRepeat', 'XSDComplexTypeClef', 'XSDComplexTypeForPart',
                                                           'XSDComplexTypeInterchangeable', 'XSDComplexTypeKey',
                                                           'XSDComplexTypeMeasureStyle',
                                                           'XSDComplexTypePartClef', 'XSDComplexTypePartTranspose', 'XSDComplexTypeSlash',
                                                           'XSDComplexTypeStaffDetails', 'XSDComplexTypeStaffTuning', 'XSDComplexTypeTime',
                                                           'XSDComplexTypeTranspose', 'XSDComplexTypeBarline', 'XSDComplexTypeAccord',
                                                           'XSDComplexTypeAccordionRegistration', 'XSDComplexTypeBass',
                                                           'XSDComplexTypeBeatUnitTied', 'XSDComplexTypeDegree', 'XSDComplexTypeDirection',
                                                           'XSDComplexTypeDirectionType', 'XSDComplexTypeFrame', 'XSDComplexTypeFrameNote',
                                                           'XSDComplexTypeGrouping', 'XSDComplexTypeHarmony', 'XSDComplexTypeHarpPedals',
                                                           'XSDComplexTypeInstrumentChange', 'XSDComplexTypeListening',
                                                           'XSDComplexTypeMetronome', 'XSDComplexTypeMetronomeNote',
                                                           'XSDComplexTypeMetronomeTuplet', 'XSDComplexTypeNumeral',
                                                           'XSDComplexTypeNumeralKey', 'XSDComplexTypePedalTuning',
                                                           'XSDComplexTypePercussion',
                                                           'XSDComplexTypePrint', 'XSDComplexTypeRoot', 'XSDComplexTypeScordatura',
                                                           'XSDComplexTypeSound', 'XSDComplexTypeStick', 'XSDComplexTypeSwing',
                                                           'XSDComplexTypeEncoding', 'XSDComplexTypeIdentification',
                                                           'XSDComplexTypeMiscellaneous', 'XSDComplexTypeAppearance',
                                                           'XSDComplexTypeMeasureLayout', 'XSDComplexTypePageLayout',
                                                           'XSDComplexTypePageMargins', 'XSDComplexTypeScaling',
                                                           'XSDComplexTypeStaffLayout',
                                                           'XSDComplexTypeSystemDividers', 'XSDComplexTypeSystemLayout',
                                                           'XSDComplexTypeSystemMargins', 'XSDComplexTypeArticulations',
                                                           'XSDComplexTypeArrow',
                                                           'XSDComplexTypeBackup', 'XSDComplexTypeBend', 'XSDComplexTypeFigure',
                                                           'XSDComplexTypeFiguredBass', 'XSDComplexTypeForward', 'XSDComplexTypeHarmonMute',
                                                           'XSDComplexTypeHarmonic', 'XSDComplexTypeHole', 'XSDComplexTypeListen',
                                                           'XSDComplexTypeLyric', 'XSDComplexTypeNotations', 'XSDComplexTypeNote',
                                                           'XSDComplexTypeNoteheadText', 'XSDComplexTypeOrnaments', 'XSDComplexTypePitch',
                                                           'XSDComplexTypeRest', 'XSDComplexTypeTechnical',
                                                           'XSDComplexTypeTimeModification',
                                                           'XSDComplexTypeTuplet', 'XSDComplexTypeTupletPortion', 'XSDComplexTypeUnpitched',
                                                           'XSDComplexTypeCredit', 'XSDComplexTypeDefaults', 'XSDComplexTypePartGroup',
                                                           'XSDComplexTypePartLink', 'XSDComplexTypePartList', 'XSDComplexTypePlayer',
                                                           'XSDComplexTypeScoreInstrument', 'XSDComplexTypeScorePart',
                                                           'XSDComplexTypeVirtualInstrument', 'XSDComplexTypeWork',
                                                           'XSDComplexTypeScorePartwise', 'XSDComplexTypePart', 'XSDComplexTypeMeasure',
                                                           }
        assert complex_types_without_child_container == ['XSDComplexTypeDirective', 'XSDComplexTypeAccidentalText', 'XSDComplexTypeCoda',
                                                         'XSDComplexTypeEmpty', 'XSDComplexTypeEmptyPlacement',
                                                         'XSDComplexTypeEmptyPlacementSmufl', 'XSDComplexTypeEmptyPrintStyle',
                                                         'XSDComplexTypeEmptyPrintStyleAlign', 'XSDComplexTypeEmptyPrintStyleAlignId',
                                                         'XSDComplexTypeEmptyPrintObjectStyleAlign', 'XSDComplexTypeEmptyTrillSound',
                                                         'XSDComplexTypeHorizontalTurn', 'XSDComplexTypeFermata', 'XSDComplexTypeFingering',
                                                         'XSDComplexTypeFormattedSymbol', 'XSDComplexTypeFormattedSymbolId',
                                                         'XSDComplexTypeFormattedText', 'XSDComplexTypeFormattedTextId',
                                                         'XSDComplexTypeFret', 'XSDComplexTypeLevel', 'XSDComplexTypeMidiDevice',
                                                         'XSDComplexTypeOtherPlay', 'XSDComplexTypeSegno', 'XSDComplexTypeString',
                                                         'XSDComplexTypeTypedText', 'XSDComplexTypeWavyLine', 'XSDComplexTypeCancel',
                                                         'XSDComplexTypeDouble', 'XSDComplexTypeKeyAccidental', 'XSDComplexTypeKeyOctave',
                                                         'XSDComplexTypeLineDetail', 'XSDComplexTypeMeasureRepeat',
                                                         'XSDComplexTypeMultipleRest', 'XSDComplexTypePartSymbol',
                                                         'XSDComplexTypeStaffSize', 'XSDComplexTypeBarStyleColor', 'XSDComplexTypeEnding',
                                                         'XSDComplexTypeRepeat', 'XSDComplexTypeBarre', 'XSDComplexTypeHarmonyAlter',
                                                         'XSDComplexTypeBassStep', 'XSDComplexTypeBeater', 'XSDComplexTypeBracket',
                                                         'XSDComplexTypeDashes', 'XSDComplexTypeDegreeAlter', 'XSDComplexTypeDegreeType',
                                                         'XSDComplexTypeDegreeValue', 'XSDComplexTypeEffect', 'XSDComplexTypeFeature',
                                                         'XSDComplexTypeFirstFret', 'XSDComplexTypeGlass', 'XSDComplexTypeImage',
                                                         'XSDComplexTypeInversion', 'XSDComplexTypeKind', 'XSDComplexTypeMeasureNumbering',
                                                         'XSDComplexTypeMembrane', 'XSDComplexTypeMetal', 'XSDComplexTypeMetronomeBeam',
                                                         'XSDComplexTypeMetronomeTied', 'XSDComplexTypeNumeralRoot',
                                                         'XSDComplexTypeOctaveShift', 'XSDComplexTypeOffset',
                                                         'XSDComplexTypeOtherDirection', 'XSDComplexTypeOtherListening',
                                                         'XSDComplexTypePedal', 'XSDComplexTypePerMinute', 'XSDComplexTypePitched',
                                                         'XSDComplexTypePrincipalVoice', 'XSDComplexTypeRootStep',
                                                         'XSDComplexTypeStaffDivide', 'XSDComplexTypeStringMute', 'XSDComplexTypeSync',
                                                         'XSDComplexTypeTimpani', 'XSDComplexTypeWedge', 'XSDComplexTypeWood',
                                                         'XSDComplexTypeMiscellaneousField', 'XSDComplexTypeSupports',
                                                         'XSDComplexTypeDistance', 'XSDComplexTypeGlyph', 'XSDComplexTypeLineWidth',
                                                         'XSDComplexTypeNoteSize', 'XSDComplexTypeOtherAppearance',
                                                         'XSDComplexTypeBookmark', 'XSDComplexTypeLink', 'XSDComplexTypeAccidental',
                                                         'XSDComplexTypeAccidentalMark', 'XSDComplexTypeArpeggiate', 'XSDComplexTypeAssess',
                                                         'XSDComplexTypeBeam', 'XSDComplexTypeBreathMark', 'XSDComplexTypeCaesura',
                                                         'XSDComplexTypeElision', 'XSDComplexTypeEmptyLine', 'XSDComplexTypeExtend',
                                                         'XSDComplexTypeGlissando', 'XSDComplexTypeGrace', 'XSDComplexTypeHammerOnPullOff',
                                                         'XSDComplexTypeHandbell', 'XSDComplexTypeHarmonClosed', 'XSDComplexTypeHeelToe',
                                                         'XSDComplexTypeHoleClosed', 'XSDComplexTypeInstrument', 'XSDComplexTypeMordent',
                                                         'XSDComplexTypeNonArpeggiate', 'XSDComplexTypeNoteType', 'XSDComplexTypeNotehead',
                                                         'XSDComplexTypeOtherNotation', 'XSDComplexTypeOtherPlacementText',
                                                         'XSDComplexTypeOtherText', 'XSDComplexTypePlacementText', 'XSDComplexTypeRelease',
                                                         'XSDComplexTypeSlide', 'XSDComplexTypeSlur', 'XSDComplexTypeStem',
                                                         'XSDComplexTypeStrongAccent', 'XSDComplexTypeStyleText', 'XSDComplexTypeTap',
                                                         'XSDComplexTypeTextElementData', 'XSDComplexTypeTie', 'XSDComplexTypeTied',
                                                         'XSDComplexTypeTremolo', 'XSDComplexTypeTupletDot', 'XSDComplexTypeTupletNumber',
                                                         'XSDComplexTypeTupletType', 'XSDComplexTypeWait', 'XSDComplexTypeEmptyFont',
                                                         'XSDComplexTypeGroupBarline', 'XSDComplexTypeGroupName',
                                                         'XSDComplexTypeGroupSymbol', 'XSDComplexTypeInstrumentLink',
                                                         'XSDComplexTypeLyricFont', 'XSDComplexTypeLyricLanguage', 'XSDComplexTypeOpus',
                                                         'XSDComplexTypePartName']

    def test_complex_type_with_complex_content_child_container(self):
        """
        complexType@name=metronome-tuplet
            complexContent
                extension@base=time-modification
                    attribute@name=type@type=start-stop@use=required
                    attribute@name=bracket@type=yes-no
                    attribute@name=show-number@type=show-tuplet
        """
        container = XMLChildContainerFactory(XSDComplexTypeMetronomeTuplet).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=actual-notes@minOccurs=1@maxOccurs=1
    ├── Element@name=normal-notes@minOccurs=1@maxOccurs=1
    └── Sequence@minOccurs=0@maxOccurs=1
        ├── Element@name=normal-type@minOccurs=1@maxOccurs=1
        └── Element@name=normal-dot@minOccurs=0@maxOccurs=unbounded
"""
        assert container.get_tree_representation() == expected

    def test_complex_type_with_group_child_container(self):
        """
        complexType@name=system-margins
            group@ref=left-right-margins
        """
        container = XMLChildContainerFactory(XSDComplexTypeSystemMargins).get_child_container()
        expected = """└── Group@name=left-right-margins@minOccurs=1@maxOccurs=1
    └── Sequence@minOccurs=1@maxOccurs=1
        ├── Element@name=left-margin@minOccurs=1@maxOccurs=1
        └── Element@name=right-margin@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

    def test_complex_type_measure_child_container(self):
        """
        element@name=measure@maxOccurs=unbounded
            complexType
                group@ref=music-data
                attributeGroup@ref=measure-attributes
        """
        container = XMLChildContainerFactory(XSDComplexTypeMeasure).get_child_container()
        expected = """└── Group@name=music-data@minOccurs=1@maxOccurs=1
    └── Sequence@minOccurs=1@maxOccurs=1
        └── Choice@minOccurs=0@maxOccurs=unbounded
            ├── Element@name=note@minOccurs=1@maxOccurs=1
            ├── Element@name=backup@minOccurs=1@maxOccurs=1
            ├── Element@name=forward@minOccurs=1@maxOccurs=1
            ├── Element@name=direction@minOccurs=1@maxOccurs=1
            ├── Element@name=attributes@minOccurs=1@maxOccurs=1
            ├── Element@name=harmony@minOccurs=1@maxOccurs=1
            ├── Element@name=figured-bass@minOccurs=1@maxOccurs=1
            ├── Element@name=print@minOccurs=1@maxOccurs=1
            ├── Element@name=sound@minOccurs=1@maxOccurs=1
            ├── Element@name=listening@minOccurs=1@maxOccurs=1
            ├── Element@name=barline@minOccurs=1@maxOccurs=1
            ├── Element@name=grouping@minOccurs=1@maxOccurs=1
            ├── Element@name=link@minOccurs=1@maxOccurs=1
            └── Element@name=bookmark@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

    def test_children_container_compact_repr(self):
        """
        Test that children container has a compact_repr attribute
        """
        container = XMLChildContainer(self.element, min_occurrences=0)
        assert container.compact_repr == 'Element@name=alter@minOccurs=0@maxOccurs=1'
        container = XMLChildContainer(self.sequence)
        assert container.compact_repr == 'Sequence@minOccurs=1@maxOccurs=1'
        container = XMLChildContainer(self.choice, min_occurrences=0, max_occurrences='unbounded')
        assert container.compact_repr == 'Choice@minOccurs=0@maxOccurs=unbounded'
        container = XMLChildContainer(XSDGroupEditorial(), min_occurrences=1, max_occurrences=2)
        assert container.compact_repr == 'Group@name=editorial@minOccurs=1@maxOccurs=2'

    def test_child_container_simple_sequence(self):
        """
        Test a complex type with a simple sequence of elements
        """
        container = XMLChildContainerFactory(complex_type=XSDComplexTypePitch).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=step@minOccurs=1@maxOccurs=1
    ├── Element@name=alter@minOccurs=0@maxOccurs=1
    └── Element@name=octave@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

        container.add_element(XMLStep('A'))
        container.add_element(XMLOctave(2))

        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=step@minOccurs=1@maxOccurs=1
        XMLStep
    ├── Element@name=alter@minOccurs=0@maxOccurs=1
    └── Element@name=octave@minOccurs=1@maxOccurs=1
        XMLOctave
"""
        assert container.get_tree_representation() == expected
        with self.assertRaises(XMLChildContainerWrongElementError):
            container.add_element(XMLAccent())
        with self.assertRaises(XMLChildContainerMaxOccursError):
            container.add_element(XMLStep('B'))

    def test_child_container_sequence_with_sequence(self):
        """
        Test a complex type with a sequence which has elements and sequences
        """
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeStaffDetails).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=staff-type@minOccurs=0@maxOccurs=1
    ├── Sequence@minOccurs=0@maxOccurs=1
    │   ├── Element@name=staff-lines@minOccurs=1@maxOccurs=1
    │   └── Element@name=line-detail@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=staff-tuning@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=capo@minOccurs=0@maxOccurs=1
    └── Element@name=staff-size@minOccurs=0@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

    def test_child_container_sequence_with_group(self):
        """
        Test a complex type with a sequence of groups an elements
        """

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeInterchangeable).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=time-relation@minOccurs=0@maxOccurs=1
    └── Group@name=time-signature@minOccurs=1@maxOccurs=unbounded
        └── Sequence@minOccurs=1@maxOccurs=1
            ├── Element@name=beats@minOccurs=1@maxOccurs=1
            └── Element@name=beat-type@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

    def test_child_container_sequence_with_choice(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeKey).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Choice@minOccurs=1@maxOccurs=1
    │   ├── Group@name=traditional-key@minOccurs=1@maxOccurs=1
    │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │       ├── Element@name=cancel@minOccurs=0@maxOccurs=1
    │   │       ├── Element@name=fifths@minOccurs=1@maxOccurs=1
    │   │       └── Element@name=mode@minOccurs=0@maxOccurs=1
    │   └── Group@name=non-traditional-key@minOccurs=0@maxOccurs=unbounded
    │       └── Sequence@minOccurs=1@maxOccurs=1
    │           ├── Element@name=key-step@minOccurs=1@maxOccurs=1
    │           ├── Element@name=key-alter@minOccurs=1@maxOccurs=1
    │           └── Element@name=key-accidental@minOccurs=0@maxOccurs=1
    └── Element@name=key-octave@minOccurs=0@maxOccurs=unbounded
"""
        assert container.get_tree_representation() == expected

    def test_child_container_complex_sequence_with_group_and_choice(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeAttributes).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Group@name=editorial@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Group@name=footnote@minOccurs=0@maxOccurs=1
    │       │   └── Sequence@minOccurs=1@maxOccurs=1
    │       │       └── Element@name=footnote@minOccurs=1@maxOccurs=1
    │       └── Group@name=level@minOccurs=0@maxOccurs=1
    │           └── Sequence@minOccurs=1@maxOccurs=1
    │               └── Element@name=level@minOccurs=1@maxOccurs=1
    ├── Element@name=divisions@minOccurs=0@maxOccurs=1
    ├── Element@name=key@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=time@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=staves@minOccurs=0@maxOccurs=1
    ├── Element@name=part-symbol@minOccurs=0@maxOccurs=1
    ├── Element@name=instruments@minOccurs=0@maxOccurs=1
    ├── Element@name=clef@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=staff-details@minOccurs=0@maxOccurs=unbounded
    ├── Choice@minOccurs=1@maxOccurs=1
    │   ├── Element@name=transpose@minOccurs=0@maxOccurs=unbounded
    │   └── Element@name=for-part@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=directive@minOccurs=0@maxOccurs=unbounded
    └── Element@name=measure-style@minOccurs=0@maxOccurs=unbounded
"""
        assert container.get_tree_representation() == expected

    def test_child_container_simple_choice(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasureStyle).get_child_container()
        expected = """└── Choice@minOccurs=1@maxOccurs=1
    ├── Element@name=multiple-rest@minOccurs=1@maxOccurs=1
    ├── Element@name=measure-repeat@minOccurs=1@maxOccurs=1
    ├── Element@name=beat-repeat@minOccurs=1@maxOccurs=1
    └── Element@name=slash@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

    def test_child_container_simple_choice_adding_child(self):
        """
        Test a complex type with a simple choice of elements can manage children.
        """
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasureStyle).get_child_container()
        container.add_element(XMLMultipleRest(1))

        expected = """└── Choice@minOccurs=1@maxOccurs=1
    ├── Element@name=multiple-rest@minOccurs=1@maxOccurs=1
        XMLMultipleRest
    ├── Element@name=measure-repeat@minOccurs=1@maxOccurs=1
    ├── Element@name=beat-repeat@minOccurs=1@maxOccurs=1
    └── Element@name=slash@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected
        with self.assertRaises(XMLChildContainerWrongElementError):
            container.add_element(XMLAccent())
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLBeatRepeat())

    def test_score_partwise_child_container(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeScorePartwise).get_child_container()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Group@name=score-header@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Element@name=work@minOccurs=0@maxOccurs=1
    │       ├── Element@name=movement-number@minOccurs=0@maxOccurs=1
    │       ├── Element@name=movement-title@minOccurs=0@maxOccurs=1
    │       ├── Element@name=identification@minOccurs=0@maxOccurs=1
    │       ├── Element@name=defaults@minOccurs=0@maxOccurs=1
    │       ├── Element@name=credit@minOccurs=0@maxOccurs=unbounded
    │       └── Element@name=part-list@minOccurs=1@maxOccurs=1
    └── Element@name=part@minOccurs=1@maxOccurs=unbounded
"""
        assert container.get_tree_representation() == expected
        container.add_element(XMLPartList())

        with self.assertRaises(XMLChildContainerMaxOccursError):
            container.add_element(XMLPartList())
        container.add_element(XMLPart())
        container.add_element(XMLPart())
        container.check_required_elements()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Group@name=score-header@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Element@name=work@minOccurs=0@maxOccurs=1
    │       ├── Element@name=movement-number@minOccurs=0@maxOccurs=1
    │       ├── Element@name=movement-title@minOccurs=0@maxOccurs=1
    │       ├── Element@name=identification@minOccurs=0@maxOccurs=1
    │       ├── Element@name=defaults@minOccurs=0@maxOccurs=1
    │       ├── Element@name=credit@minOccurs=0@maxOccurs=unbounded
    │       └── Element@name=part-list@minOccurs=1@maxOccurs=1
                XMLPartList
    └── Element@name=part@minOccurs=1@maxOccurs=unbounded
        XMLPart
        XMLPart
"""
        assert container.get_tree_representation() == expected

    def test_part_and_measure_child_container(self):
        part_expected = """└── Sequence@minOccurs=1@maxOccurs=1
    └── Element@name=measure@minOccurs=1@maxOccurs=unbounded
"""
        measure_expected = """└── Group@name=music-data@minOccurs=1@maxOccurs=1
    └── Sequence@minOccurs=1@maxOccurs=1
        └── Choice@minOccurs=0@maxOccurs=unbounded
            ├── Element@name=note@minOccurs=1@maxOccurs=1
            ├── Element@name=backup@minOccurs=1@maxOccurs=1
            ├── Element@name=forward@minOccurs=1@maxOccurs=1
            ├── Element@name=direction@minOccurs=1@maxOccurs=1
            ├── Element@name=attributes@minOccurs=1@maxOccurs=1
            ├── Element@name=harmony@minOccurs=1@maxOccurs=1
            ├── Element@name=figured-bass@minOccurs=1@maxOccurs=1
            ├── Element@name=print@minOccurs=1@maxOccurs=1
            ├── Element@name=sound@minOccurs=1@maxOccurs=1
            ├── Element@name=listening@minOccurs=1@maxOccurs=1
            ├── Element@name=barline@minOccurs=1@maxOccurs=1
            ├── Element@name=grouping@minOccurs=1@maxOccurs=1
            ├── Element@name=link@minOccurs=1@maxOccurs=1
            └── Element@name=bookmark@minOccurs=1@maxOccurs=1
"""
        part_container = XMLChildContainerFactory(complex_type=XSDComplexTypePart).get_child_container()
        measure_container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasure).get_child_container()
        assert part_container.get_tree_representation() == part_expected
        assert measure_container.get_tree_representation() == measure_expected

    def test_metronome_child_container(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMetronome).get_child_container()
        expected = """└── Choice@minOccurs=1@maxOccurs=1
    ├── Sequence@minOccurs=1@maxOccurs=1
    │   ├── Group@name=beat-unit@minOccurs=1@maxOccurs=1
    │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │       ├── Element@name=beat-unit@minOccurs=1@maxOccurs=1
    │   │       └── Element@name=beat-unit-dot@minOccurs=0@maxOccurs=unbounded
    │   ├── Element@name=beat-unit-tied@minOccurs=0@maxOccurs=unbounded
    │   └── Choice@minOccurs=1@maxOccurs=1
    │       ├── Element@name=per-minute@minOccurs=1@maxOccurs=1
    │       └── Sequence@minOccurs=1@maxOccurs=1
    │           ├── Group@name=beat-unit@minOccurs=1@maxOccurs=1
    │           │   └── Sequence@minOccurs=1@maxOccurs=1
    │           │       ├── Element@name=beat-unit@minOccurs=1@maxOccurs=1
    │           │       └── Element@name=beat-unit-dot@minOccurs=0@maxOccurs=unbounded
    │           └── Element@name=beat-unit-tied@minOccurs=0@maxOccurs=unbounded
    └── Sequence@minOccurs=1@maxOccurs=1
        ├── Element@name=metronome-arrows@minOccurs=0@maxOccurs=1
        ├── Element@name=metronome-note@minOccurs=1@maxOccurs=unbounded
        └── Sequence@minOccurs=0@maxOccurs=1
            ├── Element@name=metronome-relation@minOccurs=1@maxOccurs=1
            └── Element@name=metronome-note@minOccurs=1@maxOccurs=unbounded
"""
        assert container.get_tree_representation() == expected

    def test_add_child_lyric(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        container.check_required_elements()
        choice = container.get_children()[0]
        assert choice.requirements_fulfilled is False
        container.add_element(XMLHumming())
        assert choice.requirements_fulfilled is True
        assert choice.chosen_child == choice.get_children()[3]
        with self.assertRaises(XMLChildContainerMaxOccursError):
            container.add_element(XMLHumming())
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLLaughing())
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLText('huhu'))
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLSyllabic('begin'))

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        choice = container.get_children()[0]
        container.check_required_elements()
        container.add_element(XMLElision('something'))
        assert choice.chosen_child == choice.get_children()[0]
        container.add_element(XMLText('something'))
        container.add_element(XMLText('something'))
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLLaughing())

        container.add_element(XMLText('something'))
        container.add_element(XMLSyllabic('end'), forward=1)
        expected = """    │   ├── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   ├── Element@name=syllabic@minOccurs=0@maxOccurs=1
    │   │   ├── Element@name=text@minOccurs=1@maxOccurs=1
                XMLText
    │   │   ├── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   ├── Sequence@minOccurs=0@maxOccurs=unbounded: !!!FORCED!!!
    │   │   │   ├── Sequence@minOccurs=0@maxOccurs=1: !!!FORCED!!!
    │   │   │   │   ├── Element@name=elision@minOccurs=1@maxOccurs=1
                            XMLElision
    │   │   │   │   └── Element@name=syllabic@minOccurs=0@maxOccurs=1
                            XMLSyllabic
    │   │   │   └── Element@name=text@minOccurs=1@maxOccurs=1
                        XMLText
    │   │   │   └── Sequence@minOccurs=0@maxOccurs=unbounded: !!!FORCED!!!
    │   │   │       ├── Sequence@minOccurs=0@maxOccurs=1
    │   │   │       │   ├── Element@name=elision@minOccurs=1@maxOccurs=1
    │   │   │       │   └── Element@name=syllabic@minOccurs=0@maxOccurs=1
    │   │   │       └── Element@name=text@minOccurs=1@maxOccurs=1
                        XMLText
    │   │   └── Element@name=extend@minOccurs=0@maxOccurs=1
"""
        assert choice.chosen_child.get_tree_representation(show_force_valid) == expected

    def test_add_child_note(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()
        container.check_required_elements()
        choice = container.get_children()[0]
        selected = container.add_element(XMLPitch())
        assert selected.get_parent().chosen_child == selected
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Choice@minOccurs=1@maxOccurs=1: !!Chosen Child!!
    │   ├── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │   │   │       └── Choice@minOccurs=1@maxOccurs=1: !!Chosen Child!!
    │   │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
                            XMLPitch
    │   │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │   │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=duration@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   │       └── Element@name=duration@minOccurs=1@maxOccurs=1
                        !Required!
    │   │   └── Element@name=tie@minOccurs=0@maxOccurs=2
    │   ├── Sequence@minOccurs=1@maxOccurs=1
    │   │   ├── Element@name=cue@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │   │   │       └── Choice@minOccurs=1@maxOccurs=1
    │   │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │   │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │   │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │   │   └── Group@name=duration@minOccurs=1@maxOccurs=1
    │   │       └── Sequence@minOccurs=1@maxOccurs=1
    │   │           └── Element@name=duration@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Element@name=grace@minOccurs=1@maxOccurs=1
    │       └── Choice@minOccurs=1@maxOccurs=1
    │           ├── Sequence@minOccurs=1@maxOccurs=1
    │           │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │           │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │           │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │           │   │       └── Choice@minOccurs=1@maxOccurs=1
    │           │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │           │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │           │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │           │   └── Element@name=tie@minOccurs=0@maxOccurs=2
    │           └── Sequence@minOccurs=1@maxOccurs=1
    │               ├── Element@name=cue@minOccurs=1@maxOccurs=1
    │               └── Group@name=full-note@minOccurs=1@maxOccurs=1
    │                   └── Sequence@minOccurs=1@maxOccurs=1
    │                       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │                       └── Choice@minOccurs=1@maxOccurs=1
    │                           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │                           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │                           └── Element@name=rest@minOccurs=1@maxOccurs=1
    ├── Element@name=instrument@minOccurs=0@maxOccurs=unbounded
    ├── Group@name=editorial-voice@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Group@name=footnote@minOccurs=0@maxOccurs=1
    │       │   └── Sequence@minOccurs=1@maxOccurs=1
    │       │       └── Element@name=footnote@minOccurs=1@maxOccurs=1
    │       ├── Group@name=level@minOccurs=0@maxOccurs=1
    │       │   └── Sequence@minOccurs=1@maxOccurs=1
    │       │       └── Element@name=level@minOccurs=1@maxOccurs=1
    │       └── Group@name=voice@minOccurs=0@maxOccurs=1
    │           └── Sequence@minOccurs=1@maxOccurs=1
    │               └── Element@name=voice@minOccurs=1@maxOccurs=1
    ├── Element@name=type@minOccurs=0@maxOccurs=1
    ├── Element@name=dot@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=accidental@minOccurs=0@maxOccurs=1
    ├── Element@name=time-modification@minOccurs=0@maxOccurs=1
    ├── Element@name=stem@minOccurs=0@maxOccurs=1
    ├── Element@name=notehead@minOccurs=0@maxOccurs=1
    ├── Element@name=notehead-text@minOccurs=0@maxOccurs=1
    ├── Group@name=staff@minOccurs=0@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       └── Element@name=staff@minOccurs=1@maxOccurs=1
    ├── Element@name=beam@minOccurs=0@maxOccurs=8
    ├── Element@name=notations@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=lyric@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=play@minOccurs=0@maxOccurs=1
    └── Element@name=listen@minOccurs=0@maxOccurs=1
"""
        container.check_required_elements()
        assert container.get_tree_representation(key=show_force_valid) == expected
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLUnpitched())
        container.add_element(XMLChord())
        container.check_required_elements()
        expected = """    │   ├── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
                        XMLChord
    │   │   │       └── Choice@minOccurs=1@maxOccurs=1: !!Chosen Child!!
    │   │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
                            XMLPitch
    │   │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │   │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=duration@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   │       └── Element@name=duration@minOccurs=1@maxOccurs=1
                        !Required!
    │   │   └── Element@name=tie@minOccurs=0@maxOccurs=2
"""
        assert choice.chosen_child.get_tree_representation(key=show_force_valid) == expected
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLRest())
        container.add_element(XMLDuration(2))
        container.check_required_elements()
        expected = """    │   ├── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
                        XMLChord
    │   │   │       └── Choice@minOccurs=1@maxOccurs=1: !!Chosen Child!!
    │   │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
                            XMLPitch
    │   │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │   │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=duration@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │   │       └── Element@name=duration@minOccurs=1@maxOccurs=1
                        XMLDuration
    │   │   └── Element@name=tie@minOccurs=0@maxOccurs=2
"""
        assert choice.chosen_child.get_tree_representation(key=show_force_valid) == expected
        assert container.get_children()[0].requirements_fulfilled is True
        container.add_element(XMLVoice('1'))
        assert container.get_children()[0].requirements_fulfilled is True

    def test_container_with_unbounded_choice(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeDynamics).get_child_container()
        assert container.min_occurrences == 0
        assert container.max_occurrences == 'unbounded'

    def test_container_part_list(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypePartList).get_child_container()
        container.check_required_elements()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Group@name=part-group@minOccurs=0@maxOccurs=unbounded
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       └── Element@name=part-group@minOccurs=1@maxOccurs=1
    ├── Group@name=score-part@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       └── Element@name=score-part@minOccurs=1@maxOccurs=1
                !Required!
    └── Choice@minOccurs=0@maxOccurs=unbounded
        ├── Group@name=part-group@minOccurs=1@maxOccurs=1
        │   └── Sequence@minOccurs=1@maxOccurs=1
        │       └── Element@name=part-group@minOccurs=1@maxOccurs=1
        └── Group@name=score-part@minOccurs=1@maxOccurs=1
            └── Sequence@minOccurs=1@maxOccurs=1
                └── Element@name=score-part@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation(show_force_valid) == expected
        container.check_required_elements()
        element = XMLScorePart()
        container.add_element(element)

        expected = """└── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    ├── Group@name=part-group@minOccurs=0@maxOccurs=unbounded
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       └── Element@name=part-group@minOccurs=1@maxOccurs=1
    ├── Group@name=score-part@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │       └── Element@name=score-part@minOccurs=1@maxOccurs=1
                XMLScorePart
    └── Choice@minOccurs=0@maxOccurs=unbounded
        ├── Group@name=part-group@minOccurs=1@maxOccurs=1
        │   └── Sequence@minOccurs=1@maxOccurs=1
        │       └── Element@name=part-group@minOccurs=1@maxOccurs=1
        └── Group@name=score-part@minOccurs=1@maxOccurs=1
            └── Sequence@minOccurs=1@maxOccurs=1
                └── Element@name=score-part@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation(show_force_valid) == expected

    def test_xsd_element_parent_content(self):
        """
        Test that if an xsd element is content of a child_container it can return parent_container.
        """

        container = XMLChildContainer(content=self.element)
        assert container.content.parent_container == container
        container = XMLChildContainer(content=self.sequence)
        for node in container.traverse():
            assert node.content.parent_container == node
        container = XMLChildContainer(content=self.choice)
        for node in container.traverse():
            assert node.content.parent_container == node
        container = XMLChildContainerFactory(complex_type=XSDComplexTypePitch).get_child_container()
        for node in container.traverse():
            assert node.content.parent_container == node


class TestDuplication(TestCase):
    def test_duplication_sequence(self):
        ds = DuplicationXSDSequence()
        container = XMLChildContainer(content=ds)
        child_container = XMLChildContainerFactory(complex_type=XSDComplexTypeDynamics).get_child_container()
        ch1 = container.add_child(child_container)
        ch2 = container.add_child(child_container._create_empty_copy())
        assert container.get_children() == [ch1, ch2]

    def test_create_empty_copy(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeKey).get_child_container()
        first_group = container.get_children()[0].get_children()[0]
        first_group.add_element(XMLCancel(0))
        copied_group = first_group._create_empty_copy()
        expected = """└── Group@name=traditional-key@minOccurs=1@maxOccurs=1
    └── Sequence@minOccurs=1@maxOccurs=1
        ├── Element@name=cancel@minOccurs=0@maxOccurs=1
        ├── Element@name=fifths@minOccurs=1@maxOccurs=1
        └── Element@name=mode@minOccurs=0@maxOccurs=1
"""
        assert copied_group.get_tree_representation() == expected
        first_sequence = container.get_children()[0].get_children()[0].get_children()[0]
        copied_sequence = first_sequence._create_empty_copy()
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=cancel@minOccurs=0@maxOccurs=1
    ├── Element@name=fifths@minOccurs=1@maxOccurs=1
    └── Element@name=mode@minOccurs=0@maxOccurs=1
"""
        assert copied_sequence.get_tree_representation() == expected

        choice = container.get_children()[0]
        copied_choice = choice._create_empty_copy()
        expected = """└── Choice@minOccurs=1@maxOccurs=1
    ├── Group@name=traditional-key@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Element@name=cancel@minOccurs=0@maxOccurs=1
    │       ├── Element@name=fifths@minOccurs=1@maxOccurs=1
    │       └── Element@name=mode@minOccurs=0@maxOccurs=1
    └── Group@name=non-traditional-key@minOccurs=0@maxOccurs=unbounded
        └── Sequence@minOccurs=1@maxOccurs=1
            ├── Element@name=key-step@minOccurs=1@maxOccurs=1
            ├── Element@name=key-alter@minOccurs=1@maxOccurs=1
            └── Element@name=key-accidental@minOccurs=0@maxOccurs=1
"""
        assert copied_choice.get_tree_representation() == expected

    def test_add_duplication_parent(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeKey).get_child_container()
        """
        No parent
        """
        container._add_duplication_parent()
        assert isinstance(container.get_parent().content, DuplicationXSDSequence)
        """
        With parent
        """
        group = container.get_children()[0].get_children()[1]
        parent = group.get_parent()
        index = parent.get_children().index(group)
        group._add_duplication_parent()
        assert isinstance(group.get_parent().content, DuplicationXSDSequence)
        assert group.get_parent().get_parent() == parent
        assert parent.get_children().index(group.get_parent()) == index

    def test_duplicate_node(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeKey).get_child_container()
        container.add_element(XMLKeyAccidental('sharp'))
        unbounded_group = container.get_children()[0].get_children()[1]
        duplicated = unbounded_group.duplicate()
        assert duplicated.get_parent() == unbounded_group.get_parent()
        assert container.get_leaves(lambda x: x.content.name) == [
            (['cancel', 'fifths', 'mode'], [['key-step', 'key-alter', 'key-accidental'], ['key-step', 'key-alter', 'key-accidental']]),
            'key-octave']
        container.add_element(XMLKeyAccidental('sharp'))
        other_duplicated = unbounded_group.duplicate()
        assert duplicated.get_parent() == unbounded_group.get_parent()
        assert other_duplicated.get_parent() == duplicated.get_parent() == unbounded_group.get_parent()
        assert container.get_leaves(lambda x: x.content.name) == [
            (['cancel', 'fifths', 'mode'], [['key-step', 'key-alter', 'key-accidental'], ['key-step', 'key-alter', 'key-accidental'],
                                            ['key-step', 'key-alter', 'key-accidental']]),
            'key-octave']
        container.check_required_elements()
        container.add_element(XMLKeyAlter(1), 2)
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Choice@minOccurs=1@maxOccurs=1: !!Chosen Child!!
    │   ├── Group@name=traditional-key@minOccurs=1@maxOccurs=1
    │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │       ├── Element@name=cancel@minOccurs=0@maxOccurs=1
    │   │       ├── Element@name=fifths@minOccurs=1@maxOccurs=1
    │   │       └── Element@name=mode@minOccurs=0@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   ├── Group@name=non-traditional-key@minOccurs=0@maxOccurs=unbounded
    │   │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │   │       ├── Element@name=key-step@minOccurs=1@maxOccurs=1
                        !Required!
    │   │       ├── Element@name=key-alter@minOccurs=1@maxOccurs=1
                        !Required!
    │   │       └── Element@name=key-accidental@minOccurs=0@maxOccurs=1
                        XMLKeyAccidental
    │       ├── Group@name=non-traditional-key@minOccurs=0@maxOccurs=unbounded
    │       │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │       │       ├── Element@name=key-step@minOccurs=1@maxOccurs=1
                        !Required!
    │       │       ├── Element@name=key-alter@minOccurs=1@maxOccurs=1
                        !Required!
    │       │       └── Element@name=key-accidental@minOccurs=0@maxOccurs=1
                        XMLKeyAccidental
    │       └── Group@name=non-traditional-key@minOccurs=0@maxOccurs=unbounded
    │           └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │               ├── Element@name=key-step@minOccurs=1@maxOccurs=1
                        !Required!
    │               ├── Element@name=key-alter@minOccurs=1@maxOccurs=1
                        XMLKeyAlter
    │               └── Element@name=key-accidental@minOccurs=0@maxOccurs=1
    └── Element@name=key-octave@minOccurs=0@maxOccurs=unbounded
"""

        container.check_required_elements()
        assert container.get_tree_representation(show_force_valid) == expected

    def test_duplicate_parent(self):
        """
        Test that duplicate parent method of XMLChildContainer duplicates a parent with max
        occurrences greater one if exists on the path to root if a XSDElement
        """
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeDynamics).get_child_container()
        element = container.get_children()[0]
        duplicated_parent = element._duplicate_parent_in_path()
        assert isinstance(duplicated_parent.content, XSDChoice)
        assert duplicated_parent.get_parent() == container.get_parent()

    def test_add_child_to_unbounded_choice_dynamics(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeDynamics).get_child_container()
        container.add_element(XMLPp())
        container.add_element(XMLPp())
        container.add_element(XMLPp())
        container.get_parent().add_element(XMLPp())
        assert len(container.get_parent().get_children()) == 4
        assert [leaf.content.xml_elements[0].__class__.__name__ for leaf in container.get_parent().iterate_leaves() if
                leaf.content.xml_elements] == ['XMLPp', 'XMLPp', 'XMLPp', 'XMLPp']

        container.add_element(XMLMf())
        assert len(container.get_parent().get_children()) == 5
        assert [leaf.content.xml_elements[0].__class__.__name__ for leaf in container.get_parent().iterate_leaves() if
                leaf.content.xml_elements] == ['XMLPp', 'XMLPp', 'XMLPp', 'XMLPp', 'XMLMf']

    def test_add_child_to_unbounded_choice_articulations(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeArticulations).get_child_container()
        accent = XMLAccent()
        staccato = XMLStaccato()
        container.add_element(accent)
        container.add_element(staccato)
        assert container.up.get_attached_elements() == [accent, staccato]


class TestChildContainerCheckRequired(TestCase):
    def test_get_all_leaves(self):
        func = lambda l: convert_to_xml_class_name(l.content.name)
        container = XMLChildContainerFactory(complex_type=XSDComplexTypePitch).get_child_container()
        assert container.get_leaves(func) == ['XMLStep', 'XMLAlter', 'XMLOctave']

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeStaffDetails).get_child_container()
        assert container.get_leaves(func) == ['XMLStaffType', ['XMLStaffLines', 'XMLLineDetail'], 'XMLStaffTuning', 'XMLCapo',
                                              'XMLStaffSize']

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeInterchangeable).get_child_container()
        assert container.get_leaves(func) == ['XMLTimeRelation', ['XMLBeats', 'XMLBeatType']]

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        assert container.get_leaves(func) == [
            (
                ['XMLSyllabic', 'XMLText', [['XMLElision', 'XMLSyllabic'], 'XMLText'], 'XMLExtend'],
                'XMLExtend', 'XMLLaughing', 'XMLHumming'
            ), 'XMLEndLine', 'XMLEndParagraph', ['XMLFootnote', 'XMLLevel']
        ]
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()
        assert container.get_leaves(func) == [
            (
                [['XMLChord', ('XMLPitch', 'XMLUnpitched', 'XMLRest')], 'XMLDuration', 'XMLTie'],
                ['XMLCue', ['XMLChord', ('XMLPitch', 'XMLUnpitched', 'XMLRest')], 'XMLDuration'],
                ['XMLGrace', ([['XMLChord', ('XMLPitch', 'XMLUnpitched', 'XMLRest')], 'XMLTie'],
                              ['XMLCue', ['XMLChord', ('XMLPitch', 'XMLUnpitched', 'XMLRest')]])]
            ),
            'XMLInstrument', ['XMLFootnote', 'XMLLevel', 'XMLVoice'], 'XMLType', 'XMLDot',
            'XMLAccidental', 'XMLTimeModification', 'XMLStem', 'XMLNotehead', 'XMLNoteheadText',
            'XMLStaff', 'XMLBeam', 'XMLNotations', 'XMLLyric', 'XMLPlay', 'XMLListen']

    def test_get_required_elements(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasureStyle).get_child_container()
        container.check_required_elements()
        assert container.get_required_element_names() == ('XMLMultipleRest', 'XMLMeasureRepeat', 'XMLBeatRepeat', 'XMLSlash')
        container.add_element(XMLSlash())
        assert container.get_required_element_names() is None

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNameDisplay).get_child_container()
        assert container.get_required_element_names() is None

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeKey).get_child_container()
        assert container.get_required_element_names() == ('XMLFifths', ['XMLKeyStep', 'XMLKeyAlter'])
        container.add_element(XMLFifths(0))
        assert container.get_required_element_names() is None

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeAttributes).get_child_container()
        assert container.get_required_element_names() is None

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeScorePartwise).get_child_container()
        assert container.get_required_element_names() == ['XMLPartList', 'XMLPart']
        container.add_element(XMLPart())
        assert container.get_required_element_names() == 'XMLPartList'
        container.add_element(XMLPartList())
        assert container.get_required_element_names() is None

        container = XMLChildContainerFactory(complex_type=XSDComplexTypePart).get_child_container()
        assert container.get_required_element_names() == 'XMLMeasure'
        container.add_element(XMLMeasure())
        assert container.get_required_element_names() is None

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasure).get_child_container()
        assert container.get_required_element_names() is None

    def test_lyric_required_elements(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        assert container.get_required_element_names() == ('XMLText', 'XMLExtend', 'XMLLaughing', 'XMLHumming')

        container.add_element(XMLText('something'))
        assert container.get_required_element_names() is None
        container.add_element(XMLElision('Something'))
        assert container.get_required_element_names() == 'XMLText'

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        container.add_element(XMLText('something'))
        assert container.get_required_element_names() is None
        container.add_element(XMLSyllabic('end'))
        assert container.get_required_element_names() is None
        container.add_element(XMLSyllabic('end'))
        assert container.get_required_element_names() == ['XMLElision', 'XMLText']

    def test_note_required_elements(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()

        assert container.get_required_element_names() == (
            [('XMLPitch', 'XMLUnpitched', 'XMLRest'), 'XMLDuration'], ['XMLCue', ('XMLPitch', 'XMLUnpitched', 'XMLRest'), 'XMLDuration'],
            ['XMLGrace', (('XMLPitch', 'XMLUnpitched', 'XMLRest'), ['XMLCue', ('XMLPitch', 'XMLUnpitched', 'XMLRest')])]
        )

        container.add_element(XMLPitch())

        assert container.get_required_element_names() == 'XMLDuration'

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()
        choice = container.get_children()[0]

        container.add_element(XMLGrace())
        container.check_required_elements()
        expected = """    │   └── Sequence@minOccurs=1@maxOccurs=1: !!!FORCED!!!
    │       ├── Element@name=grace@minOccurs=1@maxOccurs=1
                XMLGrace
    │       └── Choice@minOccurs=1@maxOccurs=1
                !Required!
    │           ├── Sequence@minOccurs=1@maxOccurs=1
    │           │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │           │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │           │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │           │   │       └── Choice@minOccurs=1@maxOccurs=1
    │           │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │           │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │           │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │           │   └── Element@name=tie@minOccurs=0@maxOccurs=2
    │           └── Sequence@minOccurs=1@maxOccurs=1
    │               ├── Element@name=cue@minOccurs=1@maxOccurs=1
    │               └── Group@name=full-note@minOccurs=1@maxOccurs=1
    │                   └── Sequence@minOccurs=1@maxOccurs=1
    │                       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │                       └── Choice@minOccurs=1@maxOccurs=1
    │                           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │                           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │                           └── Element@name=rest@minOccurs=1@maxOccurs=1
"""
        assert choice.chosen_child.get_tree_representation(show_force_valid) == expected
        container.add_element(XMLPitch())
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLRest())
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLCue(), intelligent_choice=False)
        with self.assertRaises(XMLChildContainerChoiceHasAnotherChosenChild):
            container.add_element(XMLDuration(1))

    def test_required_choice_not_fulfilled(self):
        """
        Test methode add_required will show in tree presentation that the element or the choice requirement is not fulfilled
        """
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasureStyle).get_child_container()
        container.requirements_fulfilled = False
        expected = """└── Choice@minOccurs=1@maxOccurs=1
    !Required!
    ├── Element@name=multiple-rest@minOccurs=1@maxOccurs=1
    ├── Element@name=measure-repeat@minOccurs=1@maxOccurs=1
    ├── Element@name=beat-repeat@minOccurs=1@maxOccurs=1
    └── Element@name=slash@minOccurs=1@maxOccurs=1
"""
        assert container.get_tree_representation() == expected

    def test_required_element_not_fulfilled(self):
        """
        Test methode add_required will show in tree presentation that the element or the choice requirement is not fulfilled
        """
        container = XMLChildContainerFactory(complex_type=XSDComplexTypePitch).get_child_container()
        container.get_children()[0].requirements_fulfilled = False
        container.get_children()[2].requirements_fulfilled = False
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Element@name=step@minOccurs=1@maxOccurs=1
        !Required!
    ├── Element@name=alter@minOccurs=0@maxOccurs=1
    └── Element@name=octave@minOccurs=1@maxOccurs=1
        !Required!
"""     
        assert container.get_tree_representation() == expected

    def test_check_choice(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeMeasureStyle).get_child_container()
        assert container.check_required_elements() is True
        assert container.requirements_fulfilled is False
        container.add_element(XMLMultipleRest(2))
        assert container.requirements_fulfilled is True
        assert container.check_required_elements() is False

    def test_check_sequence(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypePitch).get_child_container()
        assert container.check_required_elements() is True
        assert container.get_children()[0].requirements_fulfilled is False
        assert container.get_children()[2].requirements_fulfilled is False

    def test_check_score_child_container(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeScorePartwise).get_child_container()
        assert container.check_required_elements() is True
        container.add_element(XMLPartList())
        assert container.check_required_elements() is True
        container.add_element(XMLPart())
        assert container.check_required_elements() is False

    def test_check_lyric_child_container(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        assert container.check_required_elements() is True
        container.add_element(XMLLaughing())
        assert container.check_required_elements() is False

        container = XMLChildContainerFactory(complex_type=XSDComplexTypeLyric).get_child_container()
        container.add_element(XMLText('something'))
        assert container.check_required_elements() is False
        container.add_element(XMLElision('something'))
        assert container.check_required_elements() is True
        container.add_element(XMLText('something'))
        assert container.check_required_elements() is False

    #     def test_check_intelligent_choice(self):
    #         container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()
    #         container.add_element(XMLPitch())
    #         container.add_element(XMLDuration(1))
    #         container.add_element(XMLCue())
    #         assert container.check_required_elements() is False
    #         expected = """Sequence@minOccurs=1@maxOccurs=1
    #     Choice@minOccurs=1@maxOccurs=1
    #         Sequence@minOccurs=1@maxOccurs=1
    #             Group@name=full-note@minOccurs=1@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=chord@minOccurs=0@maxOccurs=1
    #                     Choice@minOccurs=1@maxOccurs=1
    #                         Element@name=pitch@minOccurs=1@maxOccurs=1
    #                             XMLPitch
    #                         Element@name=unpitched@minOccurs=1@maxOccurs=1
    #                         Element@name=rest@minOccurs=1@maxOccurs=1
    #             Group@name=duration@minOccurs=1@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=duration@minOccurs=1@maxOccurs=1
    #                         XMLDuration
    #             Element@name=tie@minOccurs=0@maxOccurs=2
    #         Sequence@minOccurs=1@maxOccurs=1
    #             Element@name=cue@minOccurs=1@maxOccurs=1
    #             Group@name=full-note@minOccurs=1@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=chord@minOccurs=0@maxOccurs=1
    #                     Choice@minOccurs=1@maxOccurs=1
    #                         Element@name=pitch@minOccurs=1@maxOccurs=1
    #                         Element@name=unpitched@minOccurs=1@maxOccurs=1
    #                         Element@name=rest@minOccurs=1@maxOccurs=1
    #             Group@name=duration@minOccurs=1@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=duration@minOccurs=1@maxOccurs=1
    #         Sequence@minOccurs=1@maxOccurs=1
    #             Element@name=grace@minOccurs=1@maxOccurs=1
    #             Choice@minOccurs=1@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Group@name=full-note@minOccurs=1@maxOccurs=1
    #                         Sequence@minOccurs=1@maxOccurs=1
    #                             Element@name=chord@minOccurs=0@maxOccurs=1
    #                             Choice@minOccurs=1@maxOccurs=1
    #                                 Element@name=pitch@minOccurs=1@maxOccurs=1
    #                                 Element@name=unpitched@minOccurs=1@maxOccurs=1
    #                                 Element@name=rest@minOccurs=1@maxOccurs=1
    #                     Element@name=tie@minOccurs=0@maxOccurs=2
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=cue@minOccurs=1@maxOccurs=1
    #                     Group@name=full-note@minOccurs=1@maxOccurs=1
    #                         Sequence@minOccurs=1@maxOccurs=1
    #                             Element@name=chord@minOccurs=0@maxOccurs=1
    #                             Choice@minOccurs=1@maxOccurs=1
    #                                 Element@name=pitch@minOccurs=1@maxOccurs=1
    #                                 Element@name=unpitched@minOccurs=1@maxOccurs=1
    #                                 Element@name=rest@minOccurs=1@maxOccurs=1
    #     Element@name=instrument@minOccurs=0@maxOccurs=unbounded
    #     Group@name=editorial-voice@minOccurs=1@maxOccurs=1
    #         Sequence@minOccurs=1@maxOccurs=1
    #             Group@name=footnote@minOccurs=0@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=footnote@minOccurs=1@maxOccurs=1
    #             Group@name=level@minOccurs=0@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=level@minOccurs=1@maxOccurs=1
    #             Group@name=voice@minOccurs=0@maxOccurs=1
    #                 Sequence@minOccurs=1@maxOccurs=1
    #                     Element@name=voice@minOccurs=1@maxOccurs=1
    #     Element@name=type@minOccurs=0@maxOccurs=1
    #     Element@name=dot@minOccurs=0@maxOccurs=unbounded
    #     Element@name=accidental@minOccurs=0@maxOccurs=1
    #     Element@name=time-modification@minOccurs=0@maxOccurs=1
    #     Element@name=stem@minOccurs=0@maxOccurs=1
    #     Element@name=notehead@minOccurs=0@maxOccurs=1
    #     Element@name=notehead-text@minOccurs=0@maxOccurs=1
    #     Group@name=staff@minOccurs=0@maxOccurs=1
    #         Sequence@minOccurs=1@maxOccurs=1
    #             Element@name=staff@minOccurs=1@maxOccurs=1
    #     Element@name=beam@minOccurs=0@maxOccurs=8
    #     Element@name=notations@minOccurs=0@maxOccurs=unbounded
    #     Element@name=lyric@minOccurs=0@maxOccurs=unbounded
    #     Element@name=play@minOccurs=0@maxOccurs=1
    #     Element@name=listen@minOccurs=0@maxOccurs=1
    # """
    #
    #         print(container.get_tree_representation())
    #         assert container.get_tree_representation() == expected

    def test_note_with_voice_and_type(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()
        container.add_element(XMLPitch())
        container.add_element(XMLDuration(1))
        container.add_element(XMLVoice('1'))
        container.add_element(XMLType('eighth'))
        expected = """└── Sequence@minOccurs=1@maxOccurs=1
    ├── Choice@minOccurs=1@maxOccurs=1
    │   ├── Sequence@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │   │   │       └── Choice@minOccurs=1@maxOccurs=1
    │   │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
                            XMLPitch
    │   │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │   │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=duration@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │   │       └── Element@name=duration@minOccurs=1@maxOccurs=1
                        XMLDuration
    │   │   └── Element@name=tie@minOccurs=0@maxOccurs=2
    │   ├── Sequence@minOccurs=1@maxOccurs=1
    │   │   ├── Element@name=cue@minOccurs=1@maxOccurs=1
    │   │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │   │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │   │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │   │   │       └── Choice@minOccurs=1@maxOccurs=1
    │   │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │   │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │   │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │   │   └── Group@name=duration@minOccurs=1@maxOccurs=1
    │   │       └── Sequence@minOccurs=1@maxOccurs=1
    │   │           └── Element@name=duration@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Element@name=grace@minOccurs=1@maxOccurs=1
    │       └── Choice@minOccurs=1@maxOccurs=1
    │           ├── Sequence@minOccurs=1@maxOccurs=1
    │           │   ├── Group@name=full-note@minOccurs=1@maxOccurs=1
    │           │   │   └── Sequence@minOccurs=1@maxOccurs=1
    │           │   │       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │           │   │       └── Choice@minOccurs=1@maxOccurs=1
    │           │   │           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │           │   │           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │           │   │           └── Element@name=rest@minOccurs=1@maxOccurs=1
    │           │   └── Element@name=tie@minOccurs=0@maxOccurs=2
    │           └── Sequence@minOccurs=1@maxOccurs=1
    │               ├── Element@name=cue@minOccurs=1@maxOccurs=1
    │               └── Group@name=full-note@minOccurs=1@maxOccurs=1
    │                   └── Sequence@minOccurs=1@maxOccurs=1
    │                       ├── Element@name=chord@minOccurs=0@maxOccurs=1
    │                       └── Choice@minOccurs=1@maxOccurs=1
    │                           ├── Element@name=pitch@minOccurs=1@maxOccurs=1
    │                           ├── Element@name=unpitched@minOccurs=1@maxOccurs=1
    │                           └── Element@name=rest@minOccurs=1@maxOccurs=1
    ├── Element@name=instrument@minOccurs=0@maxOccurs=unbounded
    ├── Group@name=editorial-voice@minOccurs=1@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       ├── Group@name=footnote@minOccurs=0@maxOccurs=1
    │       │   └── Sequence@minOccurs=1@maxOccurs=1
    │       │       └── Element@name=footnote@minOccurs=1@maxOccurs=1
    │       ├── Group@name=level@minOccurs=0@maxOccurs=1
    │       │   └── Sequence@minOccurs=1@maxOccurs=1
    │       │       └── Element@name=level@minOccurs=1@maxOccurs=1
    │       └── Group@name=voice@minOccurs=0@maxOccurs=1
    │           └── Sequence@minOccurs=1@maxOccurs=1
    │               └── Element@name=voice@minOccurs=1@maxOccurs=1
                        XMLVoice
    ├── Element@name=type@minOccurs=0@maxOccurs=1
        XMLType
    ├── Element@name=dot@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=accidental@minOccurs=0@maxOccurs=1
    ├── Element@name=time-modification@minOccurs=0@maxOccurs=1
    ├── Element@name=stem@minOccurs=0@maxOccurs=1
    ├── Element@name=notehead@minOccurs=0@maxOccurs=1
    ├── Element@name=notehead-text@minOccurs=0@maxOccurs=1
    ├── Group@name=staff@minOccurs=0@maxOccurs=1
    │   └── Sequence@minOccurs=1@maxOccurs=1
    │       └── Element@name=staff@minOccurs=1@maxOccurs=1
    ├── Element@name=beam@minOccurs=0@maxOccurs=8
    ├── Element@name=notations@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=lyric@minOccurs=0@maxOccurs=unbounded
    ├── Element@name=play@minOccurs=0@maxOccurs=1
    └── Element@name=listen@minOccurs=0@maxOccurs=1
"""     

        assert container.get_tree_representation() == expected

    def test_copy_container(self):
        container = XMLChildContainerFactory(complex_type=XSDComplexTypeNote).get_child_container()
        before_tree = container.get_tree_representation(show_force_valid)
        container.add_element(XMLPitch())
        container.add_element(XMLDuration(2))
        container.add_element(XMLVoice('1'))
        container.add_element(XMLType('eighth'))
        copied = copy.copy(container)
        assert before_tree == copied.get_tree_representation()



def cap_first(s):
    return s[0].upper() + s[1:]


def get_cleaned_token(string_value):
    output = ' '.join(partial.strip() for partial in string_value.split('\n'))
    output = ' '.join(partial.strip() for partial in output.split('\t'))
    output = ' '.join(partial.strip() for partial in output.split('\r'))
    output = ' '.join([partial.strip() for partial in output.split(' ') if partial != ''])
    return output


def convert_to_xsd_class_name(name, type_='simple_type'):
    force_simple_type = False
    try:
        name = name.split(':')[1]
        force_simple_type = True
    except IndexError:
        pass
    if force_simple_type and type_ != 'simple_type':
        raise ValueError(f"Name {name} is starting with xs: and can only be converted to simple type.")

    name = cap_first(name)
    name = ''.join([cap_first(partial) for partial in name.split('-')])
    if type_ == 'simple_type':
        name = 'XSDSimpleType' + name
    elif type_ == 'complex_type':
        name = 'XSDComplexType' + name
    elif type_ == 'group':
        name = 'XSDGroup' + name
    else:
        raise ValueError
    return name


def convert_to_xml_class_name(name: str) -> str:
    return 'XML' + ''.join([cap_first(partial) for partial in name.split('-')])


def replace_key_underline_with_hyphen(dict_):
    output = {}
    for k, v in dict_.items():
        new_key = '-'.join(k.split('_'))
        if output.get(new_key) is not None:
            raise KeyError(f"Key {new_key} already exists in dictionary.")
        output[new_key] = v
    return output


def show_force_valid(node):
    output = node.compact_repr
    if node.force_validate:
        output += ': !!!FORCED!!!'
    if node.chosen_child:
        output += f': !!Chosen Child!!'
    return output


def show_requirements_not_fulfilled(node):
    output = node.compact_repr
    if node.requirements_fulfilled:
        output += ': !!!NOT FULTILLED!!!'
    return output



from contextlib import redirect_stdout
from pathlib import Path
from string import Template
import xml.etree.ElementTree as ET

from musicxml.generate_classes.utils import xml_xsd_et_root, musicxml_xsd_et_root, get_simple_type_all_base_classes
from musicxml.xsd.xsdtree import XSDTree, XSD_TREE_DICT

default_path_1 = Path(__file__).parent / 'defaults' / 'xsdsimpletype1.py'
default_path_2 = Path(__file__).parent / 'defaults' / 'xsdsimpletype2.py'
target_path = Path(__file__).parent.parent / 'xsd' / 'xsdsimpletype.py'

template_string = """
class $class_name($base_classes):
    \"\"\"$doc\"\"\"
    _XSD_TREE = XSD_TREE_DICT['simpleType']['$name']
"""

xsd_simple_types = ['XSDSimpleType', 'XSDSimpleTypeInteger', 'XSDSimpleTypeNonNegativeInteger', 'XSDSimpleTypePositiveInteger',
                    'XSDSimpleTypeDecimal', 'XSDSimpleTypeString', 'XSDSimpleTypeToken', 'XSDSimpleTypeDate',
                    'XSDSimpleTypeNumberOrNormal', 'XSDSimpleTypePositiveIntegerOrEmpty', 'XSDSimpleTypeFontSize',
                    'XSDSimpleTypeYesNoNumber']


def simple_type_class_as_string(simple_type_):
    xsd_tree = simple_type_[1]
    name = simple_type_[0]
    class_name = xsd_tree.xsd_element_class_name
    xsd_simple_types.append(class_name)
    base_classes = get_simple_type_all_base_classes(xsd_tree)
    doc = xsd_tree.get_doc()
    if not doc:
        doc = ""
    t = Template(template_string).substitute(class_name=class_name, base_classes=', '.join(base_classes), doc=doc, name=name)
    return t


all_simple_type_et_elements = XSD_TREE_DICT['simpleType']
simple_type_elements = [item for item in all_simple_type_et_elements.items() if item[0] not in ['yes-no-number', 'font-size',
                                                                                                'number-or-normal',
                                                                                                'positive-integer-or-empty']]
with open(target_path, 'w+') as f:
    with open(default_path_1, 'r') as default_1:
        with redirect_stdout(f):
            print(default_1.read())
    with redirect_stdout(f):
        for simple_type in simple_type_elements:
            print(simple_type_class_as_string(simple_type))
    with open(default_path_2, 'r') as default_2:
        with redirect_stdout(f):
            print(default_2.read())
            print(f'__all__={xsd_simple_types}')



from pathlib import Path
from unittest import TestCase
from musicxml.xmlelement.xmlelement import *
from musicxml.xsd.xsdcomplextype import XSDComplexTypeScorePartwise
from musicxml.xsd.xsdindicator import XSDSequence

parent_folder = Path(__file__).parent


class TestScore(TestCase):

    def test_score_partwise_type(self):
        score = XMLScorePartwise()
        assert score.TYPE == XSDComplexTypeScorePartwise

    def test_xsd_complex_type_score_partwise_sequence(self):
        assert XSDComplexTypeScorePartwise.get_xsd_indicator()[0].elements == [('XMLWork', '0', '1'), ('XMLMovementNumber', '0', '1'),
                                                                               ('XMLMovementTitle', '0', '1'),
                                                                               ('XMLIdentification', '0', '1'),
                                                                               ('XMLDefaults', '0', '1'), ('XMLCredit', '0', 'unbounded'),
                                                                               ('XMLPartList', '1', '1'), ('XMLPart', '1', 'unbounded')]

    def test_score_partwise_indicator(self):
        score = XMLScorePartwise()
        assert isinstance(score.TYPE.get_xsd_indicator()[0], XSDSequence)

    def test_xsd_tree(self):
        score = XMLScorePartwise()

        expected = """└── XSDTree element@name=score-partwise@block=extension substitution@final=#all
    ├── XSDTree annotation
    │   └── XSDTree documentation
    └── XSDTree complexType
        ├── XSDTree sequence
        │   ├── XSDTree group@ref=score-header
        │   └── XSDTree element@name=part@maxOccurs=unbounded
        │       └── XSDTree complexType
        │           ├── XSDTree sequence
        │           │   └── XSDTree element@name=measure@maxOccurs=unbounded
        │           │       └── XSDTree complexType
        │           │           ├── XSDTree group@ref=music-data
        │           │           └── XSDTree attributeGroup@ref=measure-attributes
        │           └── XSDTree attributeGroup@ref=part-attributes
        └── XSDTree attributeGroup@ref=document-attributes
"""
        assert score.XSD_TREE.get_tree_representation() == expected

    def test_minimum_score(self):
        score = XMLScorePartwise()
        pl = score.add_child(XMLPartList())
        sp = pl.add_child(XMLScorePart(id='P1'))
        sp.add_child(XMLPartName('some name', print_object="no"))
        p = score.add_child(XMLPart(id='P1'))
        p.add_child(XMLMeasure(number='1'))

        score.write(parent_folder / 'test_minimum_score.xml')

    def test_hello_world(self):
        score = XMLScorePartwise(version='3.1')
        pl = score.add_child(XMLPartList())
        sp = pl.add_child(XMLScorePart(id='P1'))
        sp.add_child(XMLPartName('Part 1', print_object='no'))
        p = score.add_child(XMLPart(id='P1'))
        m = p.add_child(XMLMeasure(number='1'))
        att = m.add_child(XMLAttributes())
        att.add_child(XMLDivisions(1))
        t = att.add_child(XMLTime())
        t.add_child(XMLBeats('4'))
        t.add_child(XMLBeatType('4'))
        c = att.add_child(XMLClef())
        c.add_child(XMLSign('G'))
        c.add_child(XMLLine(2))
        k = att.add_child(XMLKey())
        k.add_child(XMLFifths(0))
        k.add_child(XMLMode('major'))
        n = m.add_child(XMLNote())
        p = n.add_child(XMLPitch())
        p.add_child(XMLStep('C'))
        p.add_child(XMLOctave(4))
        n.add_child(XMLDuration(4))
        n.add_child(XMLVoice('1'))
        n.add_child(XMLType('whole'))
        bl = m.add_child(XMLBarline(location='right'))
        bl.add_child(XMLBarStyle('light-heavy'))
        score.write(parent_folder / 'test_hello_world_actual.xml')



from unittest import TestCase

from musicxml.xsd.xsdattribute import *
from musicxml.xsd.xsdsimpletype import *
from musicxml.xsd.xsdtree import XSDTree
from xml.etree import ElementTree as ET

from musicxml.tests.util import MusicXmlTestCase


class TestXSDAttribute(TestCase):
    def test_xsd_attribute_from_XSD_TREE(self):
        """
        Test that an XSDAttribute can be created out of a XSDTree with tag attribute.
        """
        el = ET.fromstring('<xs:attribute xmlns:xs="http://www.w3.org/2001/XMLSchema" name="type" type="xs:token" />')
        XSD_TREE = XSDTree(el)
        attribute = XSDAttribute(XSD_TREE)
        assert attribute.name == 'type'
        assert attribute.type_ == XSDSimpleTypeToken
        assert attribute.is_required is False
        assert isinstance(attribute('Hi'), XSDSimpleTypeToken)
        with self.assertRaises(TypeError):
            attribute(1)

    def test_attributes_with_ref(self):
        """
        There are a few attributes with ref instead of name:
        <xs:attribute ref="xml:lang"/>
        used in:
            <xs:element name="directive">
            <xs:complexType name="text-element-data">
            <xs:complexType name="lyric-language">
            <xs:attributeGroup name="text-formatting">
        <xs:attribute ref="xml:space"/>
        used in:
            <xs:attributeGroup name="text-formatting">

        <xs:attribute ref="xlink:href" use="required"/>
        <xs:attribute ref="xlink:type" fixed="simple"/>
        <xs:attribute ref="xlink:role"/>
        <xs:attribute ref="xlink:title"/>
        <xs:attribute ref="xlink:show" default="replace"/>
        <xs:attribute ref="xlink:actuate" default="onRequest"/>
        used in:
            <xs:attributeGroup name="link-attributes">
        """
        """
        Test lang and space
        """
        tf = XSDAttributeGroupTextFormatting()
        assert [a.name for a in tf.get_xsd_attributes()] == ['justify', 'default-x', 'default-y', 'relative-x', 'relative-y', 'font-family',
                                                             'font-style', 'font-size', 'font-weight', 'color', 'halign', 'valign',
                                                             'underline', 'overline', 'line-through', 'rotation', 'letter-spacing',
                                                             'line-height', 'lang', 'space', 'dir', 'enclosure']

        """
        Test xlink
        """
    #
    # self.fail('Incomplete')


class TestXSDAttributeGroup(MusicXmlTestCase):

    def test_attribute_group_get_attributes(self):
        """
        attributeGroup@name=position

        attribute@name=default-x@type=tenths
        attribute@name=default-y@type=tenths
        attribute@name=relative-x@type=tenths
        attribute@name=relative-y@type=tenths
        """
        [attribute_1, attribute_2, attribute_3, attribute_4] = XSDAttributeGroupPosition.get_xsd_attributes()
        assert str(attribute_1) == 'XSDAttribute@name=default-x@type=tenths'
        assert str(attribute_2) == 'XSDAttribute@name=default-y@type=tenths'
        assert str(attribute_3) == 'XSDAttribute@name=relative-x@type=tenths'
        assert str(attribute_4) == 'XSDAttribute@name=relative-y@type=tenths'

        for attribute in XSDAttributeGroupPosition.get_xsd_attributes():
            assert isinstance(attribute, XSDAttribute)
            assert isinstance(attribute(10), XSDSimpleTypeTenths)

    def test_print_style_attribute_group(self):
        """
        attributeGroup@name=print-style
            attributeGroup@ref=position
            attributeGroup@ref=font
            attributeGroup@ref=color

        attributeGroup@name=position
            attribute@name=default-x@type=tenths
            attribute@name=default-y@type=tenths
            attribute@name=relative-x@type=tenths
            attribute@name=relative-y@type=tenths

        attributeGroup@name=font
            attribute@name=font-family@type=font-family
            attribute@name=font-style@type=font-style
            attribute@name=font-size@type=font-size
            attribute@name=font-weight@type=font-weight

        attributeGroup@name=color
            attribute@name=color@type=color
        """
        [attribute_1, attribute_2, attribute_3, attribute_4, attribute_5, attribute_6, attribute_7, attribute_8, attribute_9] = \
            XSDAttributeGroupPrintStyle.get_xsd_attributes()
        assert str(attribute_1) == 'XSDAttribute@name=default-x@type=tenths'
        assert str(attribute_2) == 'XSDAttribute@name=default-y@type=tenths'
        assert str(attribute_3) == 'XSDAttribute@name=relative-x@type=tenths'
        assert str(attribute_4) == 'XSDAttribute@name=relative-y@type=tenths'
        assert str(attribute_5) == 'XSDAttribute@name=font-family@type=font-family'
        assert str(attribute_6) == 'XSDAttribute@name=font-style@type=font-style'
        assert str(attribute_7) == 'XSDAttribute@name=font-size@type=font-size'
        assert str(attribute_8) == 'XSDAttribute@name=font-weight@type=font-weight'
        assert str(attribute_9) == 'XSDAttribute@name=color@type=color'



from musicxml.xsd.xsdtree import XSDTree


class XSDElement:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self.xsd_tree = xsd_tree
        self._name = None
        self._xml_elements = []
        self.parent_container = None

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'element':
            raise ValueError
        self._xsd_tree = value

    @property
    def name(self):
        if self._name is None:
            self._name = self.xsd_tree.name
        return self._name

    def add_xml_element(self, el):
        if el.name != self.name:
            raise TypeError
        el.parent_xsd_element = self
        self._xml_elements.append(el)

    @property
    def xml_elements(self):
        return self._xml_elements

    def __copy__(self):
        return self.__class__(self.xsd_tree.__deepcopy__())



from unittest import TestCase
from xml.etree import ElementTree as ET

from musicxml.generate_classes.utils import musicxml_xsd_path, ns
from musicxml.xsd.xsdtree import XSDTree


class MusicXmlTestCase(TestCase):
    def setUp(self) -> None:
        with open(musicxml_xsd_path) as file:
            xmltree = ET.parse(file)
        self.root = xmltree.getroot()
        self.all_simple_type_xsd_elements = [XSDTree(simpletype) for simpletype in
                                             self.root.findall(f"{ns}simpleType")]
        self.all_complex_type_xsd_elements = [XSDTree(complextype) for complextype in
                                              self.root.findall(f"{ns}complexType")]
        self.above_below_simple_type_xsd_element = XSDTree(self.root.find(f"{ns}simpleType["
                                                                          f"@name='above-below']"))
        self.yes_no_number_simple_type_xsd_element = XSDTree(self.root.find(f"{ns}simpleType["
                                                                            f"@name='yes-no-number']"))
        self.complex_type_xsd_element = XSDTree(self.root.find(f"{ns}complexType[@name='fingering']"))
        self.attribute_group_position = XSDTree(self.root.find(f"{ns}attributeGroup[@name='position']"))
        super().setUp()


class XMLElementException(Exception):
    pass


class XMLElementCannotHaveChildrenError(XMLElementException):
    pass


class XMLChildContainerException(Exception):
    pass


class XMLChildContainerWrongElementError(XMLChildContainerException):
    pass


class XMLChildContainerMaxOccursError(XMLChildContainerException):
    pass


class XMLChildContainerChoiceHasAnotherChosenChild(XMLChildContainerException):
    pass


class XMLChildContainerFactoryException(Exception):
    pass


class XMLChildContainerFactoryError(XMLChildContainerFactoryException):
    pass



from musicxml.util.core import convert_to_xml_class_name
import xml.etree.ElementTree as ET
from musicxml.xmlelement.xmlelement import *


def _et_xml_to_music_xml(node):
    if node.text:
        text = node.text.strip()
    else:
        text = ''

    try:
        output = eval(convert_to_xml_class_name(node.tag))(value_=text)
    except TypeError:
        try:
            output = eval(convert_to_xml_class_name(node.tag))(value_=float(text))
        except TypeError:
            output = eval(convert_to_xml_class_name(node.tag))(value_=int(text))

    for k, v in node.attrib.items():
        try:
            setattr(output, k, v)
        except (TypeError, ValueError):
            try:
                setattr(output, k, int(v))
            except ValueError:
                setattr(output, k, float(v))

    return output


def _parse_node(xml_node):
    output = _et_xml_to_music_xml(xml_node)
    for child in xml_node:
        output.add_child(_parse_node(child))
    return output


def parse_musicxml(file_path):
    with open(file_path) as file:
        xml = ET.parse(file)
    return _parse_node(xml.getroot())



import datetime
import random
from pathlib import Path
from unittest import TestCase

from musicxml.parser.parser import parse_musicxml, _parse_node
from musicxml.xmlelement.xmlelement import XMLScorePartwise
import xml.etree.ElementTree as ET


class TestParseMusicXml(TestCase):
    def test_parse_hello_world(self):
        score = parse_musicxml(Path(__file__).parent / 'test_hello_world.xml')
        assert isinstance(score, XMLScorePartwise)
        attributes = score.get_children()[1].get_children()[0].get_children()[0].child_container_tree
        assert attributes.check_required_elements() is False
        score.write(Path(__file__).parent / 'test_hello_world_recreated.xml', intelligent_choice=True)

    def test_parse_bach_partita_3_reduced(self):
        start_reading = datetime.datetime.now()
        print("start reading score")
        score = parse_musicxml(Path(__file__).parent / 'test_bach_partita_3_reduced.xml')
        start_writing = datetime.datetime.now()
        print(f"start writing score :{start_writing - start_reading}")
        score.write(Path(__file__).parent / 'test_bach_partita_3_reduced_created.xml')
        end = datetime.datetime.now()
        print(f"end writing score :{end - start_writing}")
        """
        main.diff_files takes too long ...
        """
        # diff = main.diff_files('test_bach_partita_3_reduced.xml', 'test_bach_partita_3_reduced_created.xml')
        # assert diff == []

    def test_parse_bach_partita_3_reduced_reordered(self):
        def reordered_children(el, seed=None):
            output = ET.Element(el.tag, el.attrib)
            output.text = el.text
            children = list(el)
            if el.tag in ['note']:
                random.Random(seed).shuffle(children)
            for child in children:
                output.append(reordered_children(el=child, seed=seed))
            return output

        start_reading = datetime.datetime.now()
        with open(Path(__file__).parent / 'test_bach_partita_3_reduced.xml') as file:
            xml = ET.parse(file)
        reordered = reordered_children(xml.getroot(), seed=55)
        score = _parse_node(reordered)
        start_writing = datetime.datetime.now()
        print(f"start writing score :{start_writing - start_reading}")
        score.write(Path(__file__).parent / 'test_bach_partita_3_reordered_created.xml')
        end = datetime.datetime.now()
        print(f"end writing score :{end - start_writing}")
        # print("start reading score")

    def test_parse_bach_partita_3(self):
        start_reading = datetime.datetime.now()
        print("start reading score")
        score = parse_musicxml(Path(__file__).parent / 'test_bach_partita_3.xml')
        start_writing = datetime.datetime.now()
        print(f"start writing score :{start_writing - start_reading}")
        score.write(Path(__file__).parent / 'test_bach_partita_3_created.xml')
        end = datetime.datetime.now()
        print(f"end writing score :{end - start_writing}")



from musicxml.util.core import convert_to_xsd_class_name
from musicxml.xsd.xsdtree import XSDTree, XSDTreeElement, XSD_TREE_DICT
from musicxml.xsd.xsdsimpletype import *
import xml.etree.ElementTree as ET


class XSDAttribute:
    def __init__(self, xsd_tree):
        self._name = None
        self._ref = None
        self._xsd_tree = None
        self.xsd_tree = xsd_tree
        self._type = None
        self._is_required = None

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'attribute':
            raise ValueError
        ref = value.get_attributes().get('ref')
        if ref:
            if ref == 'xml:lang':
                self._xsd_tree = XSDTree(ET.fromstring("""<xs:attribute xmlns:xs="http://www.w3.org/2001/XMLSchema" name="lang" type="xs:language">
        <xs:annotation>
            <xs:documentation>In due course, we should install the relevant ISO 2- and 3-letter
                codes as the enumerated possible values . . .
            </xs:documentation>
        </xs:annotation>
    </xs:attribute>
    """
                                                       ))
            elif ref == 'xml:space':
                self._xsd_tree = XSDTree(ET.fromstring("""<xs:attribute xmlns:xs="http://www.w3.org/2001/XMLSchema" name="space" default="preserve">
        <xs:simpleType>
            <xs:restriction base="xs:NCName">
                <xs:enumeration value="default"/>
                <xs:enumeration value="preserve"/>
            </xs:restriction>
        </xs:simpleType>
    </xs:attribute>
    """
                                                       ))
            else:
                NotImplementedError(ref)
        else:
            self._xsd_tree = value

    @property
    def name(self):
        if self._name is None:
            self._name = self.xsd_tree.get_attributes().get('name')
        return self._name

    @property
    def ref(self):
        if self._ref is None:
            self._ref = self.xsd_tree.get_attributes().get('ref')
        return self._ref

    @property
    def type_(self):
        if self._type is None:
            self._type = eval(convert_to_xsd_class_name(self.xsd_tree.get_attributes()['type'], 'simple_type'))
        return self._type

    @property
    def is_required(self):
        if self._is_required is None:
            if self.xsd_tree.get_attributes().get('use') == 'required':
                self._is_required = True
            else:
                self._is_required = False
        return self._is_required

    def __call__(self, value):
        return self.type_(value)

    def __str__(self):
        attrs = self.xsd_tree.get_attributes()
        return f"XSDAttribute{''.join([f'@{attribute}={self.xsd_tree.get_attributes()[attribute]}' for attribute in attrs])}"

    def __repr__(self):
        return self.__str__()


class XSDAttributeGroup(XSDTreeElement):
    _XSD_ATTRIBUTES = None

    @classmethod
    def get_xsd_attributes(cls):
        if cls._XSD_ATTRIBUTES is None:
            cls._XSD_ATTRIBUTES = []
            for child in cls.XSD_TREE.get_children():
                if child.tag == 'attribute':
                    cls._XSD_ATTRIBUTES.append(XSDAttribute(child))
                if child.tag == 'attributeGroup':
                    cls._XSD_ATTRIBUTES.extend(eval(child.xsd_element_class_name).get_xsd_attributes())
        return cls._XSD_ATTRIBUTES

# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_attributes.py
# -----------------------------------------------------



from musicxml.generate_classes.utils import musicxml_xsd_et_root
from musicxml.util.core import convert_to_xsd_class_name
from musicxml.xsd.xsdsimpletype import *
from musicxml.xsd.xsdattribute import *
from musicxml.xsd.xsdindicator import *
from musicxml.xsd.xsdtree import XSDTreeElement, XSDTree, XSD_TREE_DICT
import xml.etree.ElementTree as ET


class XSDComplexType(XSDTreeElement):
    _SIMPLE_CONTENT = None
    _SEARCH_FOR_ELEMENT = ''
    _XSD_ATTRIBUTES = None

    def __init__(self, value=None, parent=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent
        self.value = value
        self._value = None

    def _get_error_class(self):
        if self.parent:
            return self.parent.__class__.__name__
        else:
            return self.__class__.__name__

    def _check_value(self, val):
        if self._SIMPLE_CONTENT:
            try:
                self._SIMPLE_CONTENT(val)
            except TypeError as err:
                raise TypeError(f"{self._get_error_class()}: " + err.args[0])
            except ValueError as err:
                raise ValueError(f"{self._get_error_class()}: " + err.args[0])

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._check_value(v)
        self._value = v

    @classmethod
    def get_xsd_attributes(cls):
        if cls._XSD_ATTRIBUTES is None:
            cls._XSD_ATTRIBUTES = []
            if cls.get_xsd_tree().get_simple_content_extension():
                for child in cls.get_xsd_tree().get_simple_content_extension().get_children():
                    if child.tag == 'attribute':
                        cls._XSD_ATTRIBUTES.append(XSDAttribute(child))
                    elif child.tag == 'attributeGroup':
                        cls._XSD_ATTRIBUTES.extend(eval(child.xsd_element_class_name).get_xsd_attributes())
            elif cls.get_xsd_tree().get_complex_content():
                complex_content_extension = cls.get_xsd_tree().get_complex_content_extension()
                complex_type_extension_base_class_name = convert_to_xsd_class_name(complex_content_extension.get_attributes()['base'],
                                                                                   'complex_type')
                extension_base = eval(complex_type_extension_base_class_name)
                cls._XSD_ATTRIBUTES.extend(extension_base.get_xsd_attributes())
                for child in complex_content_extension.get_children():
                    if child.tag == 'attribute':
                        cls._XSD_ATTRIBUTES.append(XSDAttribute(child))
                    elif child.tag == 'attributeGroup':
                        cls._XSD_ATTRIBUTES.extend(eval(child.xsd_element_class_name).get_xsd_attributes())
                # return output
            else:
                for child in cls.get_xsd_tree().get_children():
                    if child.tag == 'attribute':
                        cls._XSD_ATTRIBUTES.append(XSDAttribute(child))
                    elif child.tag == 'attributeGroup':
                        cls._XSD_ATTRIBUTES.extend(eval(child.xsd_element_class_name).get_xsd_attributes())
        return cls._XSD_ATTRIBUTES

    @classmethod
    def get_xsd_indicator(cls):
        def get_occurrences(ch):
            min_ = ch.get_attributes().get('minOccurs')
            max_ = ch.get_attributes().get('maxOccurs')
            return 1 if not min_ else int(min_), 1 if not max_ else 'unbounded' if max_ == 'unbounded' else int(max_)

        for child in cls.get_xsd_tree().get_children():
            if child.tag == 'sequence':
                return XSDSequence(child), *get_occurrences(child)
            if child.tag == 'choice':
                return XSDChoice(child), *get_occurrences(child)
            if child.tag == 'group':
                return eval(convert_to_xsd_class_name(child.get_attributes()['ref'], 'group'))(), *get_occurrences(child)
            if child.tag == 'complexContent':
                return eval(convert_to_xsd_class_name(child.get_children()[0].get_attributes()['base'],
                                                      'complex_type')).get_xsd_indicator()


class XSDComplexTypeScorePartwise(XSDComplexType):
    _SEARCH_FOR_ELEMENT = ".//{*}element[@name='score-partwise']//{*}complexType"


class XSDComplexTypePart(XSDComplexType):
    _SEARCH_FOR_ELEMENT = ".//{*}element[@name='score-partwise']//{*}complexType//{*}complexType"


class XSDComplexTypeMeasure(XSDComplexType):
    _SEARCH_FOR_ELEMENT = ".//{*}element[@name='score-partwise']//{*}complexType//{*}complexType//{*}complexType"


class XSDComplexTypeDirective(XSDComplexType):
    _SIMPLE_CONTENT = XSDSimpleTypeString
    _SEARCH_FOR_ELEMENT = ".//{*}complexType[@name='attributes']//{*}complexType"


# Note's choice is being manually reordered to avoid using intelligent choice for each Note without grace.
class XSDComplexTypeNote(XSDComplexType):
    """Notes are the most common type of MusicXML data. The MusicXML format distinguishes between elements used for sound information and elements used for notation information (e.g., tie is used for sound, tied for notation). Thus grace notes do not have a duration element. Cue notes have a duration element, as do forward elements, but no tie elements. Having these two types of information available can make interchange easier, as some programs handle one type of information more readily than the other.

The print-leger attribute is used to indicate whether leger lines are printed. Notes without leger lines are used to indicate indeterminate high and low notes. By default, it is set to yes. If print-object is set to no, print-leger is interpreted to also be set to no if not present. This attribute is ignored for rests.

The dynamics and end-dynamics attributes correspond to MIDI 1.0's Note On and Note Off velocities, respectively. They are expressed in terms of percentages of the default forte value (90 for MIDI 1.0).

The attack and release attributes are used to alter the starting and stopping time of the note from when it would otherwise occur based on the flow of durations - information that is specific to a performance. They are expressed in terms of divisions, either positive or negative. A note that starts a tie should not have a release attribute, and a note that stops a tie should not have an attack attribute. The attack and release attributes are independent of each other. The attack attribute only changes the starting time of a note, and the release attribute only changes the stopping time of a note.

If a note is played only particular times through a repeat, the time-only attribute shows which times to play the note.

The pizzicato attribute is used when just this note is sounded pizzicato, vs. the pizzicato element which changes overall playback between pizzicato and arco.
"""

    _XSD_TREE = XSDTree(ET.fromstring("""
<xs:complexType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="note">
    <xs:annotation>
        <xs:documentation>Notes are the most common type of MusicXML data. The MusicXML format distinguishes between elements used for sound information and elements used for notation information (e.g., tie is used for sound, tied for notation). Thus grace notes do not have a duration element. Cue notes have a duration element, as do forward elements, but no tie elements. Having these two types of information available can make interchange easier, as some programs handle one type of information more readily than the other.

The print-leger attribute is used to indicate whether leger lines are printed. Notes without leger lines are used to indicate indeterminate high and low notes. By default, it is set to yes. If print-object is set to no, print-leger is interpreted to also be set to no if not present. This attribute is ignored for rests.

The dynamics and end-dynamics attributes correspond to MIDI 1.0's Note On and Note Off velocities, respectively. They are expressed in terms of percentages of the default forte value (90 for MIDI 1.0).

The attack and release attributes are used to alter the starting and stopping time of the note from when it would otherwise occur based on the flow of durations - information that is specific to a performance. They are expressed in terms of divisions, either positive or negative. A note that starts a tie should not have a release attribute, and a note that stops a tie should not have an attack attribute. The attack and release attributes are independent of each other. The attack attribute only changes the starting time of a note, and the release attribute only changes the stopping time of a note.

If a note is played only particular times through a repeat, the time-only attribute shows which times to play the note.

The pizzicato attribute is used when just this note is sounded pizzicato, vs. the pizzicato element which changes overall playback between pizzicato and arco.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:choice>
            <xs:sequence>
                <xs:group ref="full-note" />
                <xs:group ref="duration" />
                <xs:element name="tie" type="tie" minOccurs="0" maxOccurs="2" />
            </xs:sequence>
            <xs:sequence>
                <xs:element name="cue" type="empty">
                    <xs:annotation>
                        <xs:documentation>The cue element indicates the presence of a cue note. In MusicXML, a cue note is a silent note with no playback. Normal notes that play can be specified as cue size using the type element. A cue note that is specified as full size using the type element will still remain silent.</xs:documentation>
                    </xs:annotation>
                </xs:element>
                <xs:group ref="full-note" />
                <xs:group ref="duration" />
            </xs:sequence>
            <xs:sequence>
                <xs:element name="grace" type="grace" />
                <xs:choice>
                    <xs:sequence>
                        <xs:group ref="full-note" />
                        <xs:element name="tie" type="tie" minOccurs="0" maxOccurs="2" />
                    </xs:sequence>
                    <xs:sequence>
                        <xs:element name="cue" type="empty" />
                        <xs:group ref="full-note" />
                    </xs:sequence>
                </xs:choice>
            </xs:sequence>
        </xs:choice>
        <xs:element name="instrument" type="instrument" minOccurs="0" maxOccurs="unbounded" />
        <xs:group ref="editorial-voice" />
        <xs:element name="type" type="note-type" minOccurs="0" />
        <xs:element name="dot" type="empty-placement" minOccurs="0" maxOccurs="unbounded">
            <xs:annotation>
                <xs:documentation>One dot element is used for each dot of prolongation. The placement attribute is used to specify whether the dot should appear above or below the staff line. It is ignored for notes that appear on a staff space.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="accidental" type="accidental" minOccurs="0" />
        <xs:element name="time-modification" type="time-modification" minOccurs="0" />
        <xs:element name="stem" type="stem" minOccurs="0" />
        <xs:element name="notehead" type="notehead" minOccurs="0" />
        <xs:element name="notehead-text" type="notehead-text" minOccurs="0" />
        <xs:group ref="staff" minOccurs="0" />
        <xs:element name="beam" type="beam" minOccurs="0" maxOccurs="8" />
        <xs:element name="notations" type="notations" minOccurs="0" maxOccurs="unbounded" />
        <xs:element name="lyric" type="lyric" minOccurs="0" maxOccurs="unbounded" />
        <xs:element name="play" type="play" minOccurs="0" />
        <xs:element name="listen" type="listen" minOccurs="0" />
    </xs:sequence>
    <xs:attributeGroup ref="x-position" />
    <xs:attributeGroup ref="font" />
    <xs:attributeGroup ref="color" />
    <xs:attributeGroup ref="printout" />
    <xs:attribute name="print-leger" type="yes-no" />
    <xs:attribute name="dynamics" type="non-negative-decimal" />
    <xs:attribute name="end-dynamics" type="non-negative-decimal" />
    <xs:attribute name="attack" type="divisions" />
    <xs:attribute name="release" type="divisions" />
    <xs:attribute name="time-only" type="time-only" />
    <xs:attribute name="pizzicato" type="yes-no" />
    <xs:attributeGroup ref="optional-unique-id" />
</xs:complexType>
"""
                                      ))
# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_complex_types.py
# -----------------------------------------------------



from musicxml.xmlelement.xmlchildcontainer import XMLChildContainerFactory
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdcomplextype import __all__

containers = {}

for ct in __all__[1:]:
    cls = eval(ct)
    if cls.get_xsd_indicator():
        containers[ct] = XMLChildContainerFactory(complex_type=cls).get_child_container()
